import numpy as np
import pytest

import pynbody


@pytest.mark.parametrize("load_all", [True, False])
def test_swift_velociraptor(load_all):
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    assert pynbody.halo.velociraptor.VelociraptorCatalogue._can_load(f)
    h = pynbody.halo.velociraptor.VelociraptorCatalogue(f)
    if load_all:
        h.load_all()
    assert len(h) == 209

    assert len(h[1]) == 443
    assert len(h[1].dm) == 246
    assert len(h[1].gas) == 197



    h_unbound = pynbody.halo.velociraptor.VelociraptorCatalogue(f, include_unbound=True)
    if load_all:
        h_unbound.load_all()
    assert len(h_unbound[1]) == 444

    testvals = [395183, 411313, 394929, 386993, 419631, 402993, 411571, 395313,
          386865, 419505, 411439, 419503, 395059, 395187, 419507, 402995,
          411437, 395057, 395315, 411565, 419635, 403121, 403377, 419633,
          403253, 411443, 411441, 395185, 403123, 403251, 403371, 403507,
          395181, 403509, 419761, 411183, 403117, 411185, 419629, 411697,
          411445, 395053, 411573, 419763, 411317, 403373, 395055, 403255,
          386995, 411435, 411311, 411569, 411309, 403247, 411563, 403125,
          419759, 403245, 395061, 411567, 419500, 394922, 386608, 395050,
          394932, 403124, 411444, 403254, 395182, 411186, 402988, 411566,
          419630, 411312, 386862, 403376, 411442, 419632, 419502, 403244,
          411316, 394928, 386990, 402992, 395056, 394924, 403246, 386994,
          403122, 402990, 395060, 395186, 395054, 386864, 394800, 403374,
          386866, 395184, 403250, 395058, 403248, 386992, 403252, 411564,
          386738, 411436, 411562, 419504, 395310, 403116, 403372, 403370,
          411438, 411568, 411314, 403120, 411310, 419628, 411182, 411570,
          394930, 395314, 411440, 411308, 386736, 402994, 419634, 411306,
          386860, 411434, 395188, 403118, 395180, 394798, 419506, 411184,
          403378, 402986]

    assert (np.sort(h[20]['iord']) == np.sort(testvals)).all()


@pytest.mark.filterwarnings("ignore:Accessing multiple halos")
@pytest.mark.parametrize("use_all", [True, False])
def test_swift_velociraptor_parents_and_children(use_all):
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    h = pynbody.halo.velociraptor.VelociraptorCatalogue(f)

    if use_all:
        properties = h.get_properties_all_halos()
        getter = lambda k, i: properties[k][h.number_mapper.number_to_index(i)]
    else:
        getter = lambda k, i: h[i].properties[k]

    assert getter('parent', 2) == -1
    assert (getter('children', 2) == [208]).all()
    assert getter('parent', 208) == 2
    assert (getter('children', 3) == [209]).all()
    assert getter('parent', 209) == 3

    # the above were from the genuine velociraptor output. But since these were the only subhalos, the test was
    # a bit too trivial. The following are some faked subhalos to test a more complex scenario

    assert (getter('children', 1) == [203, 204, 206]).all()
    assert (getter('children', 4) == [205, 207]).all()
    assert getter('parent', 203) == 1

@pytest.mark.filterwarnings("ignore:Accessing multiple halos")
@pytest.mark.parametrize("load_all", [True, False])
def test_swift_velociraptor_properties(load_all):
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    h = pynbody.halo.velociraptor.VelociraptorCatalogue(f)
    if load_all:
        h.load_all()

    assert np.allclose(float(h[1].properties['Lx']), -90787.3983020414e13)
    assert np.allclose(float(h[10].properties['Lx']), -73881.83861892551e13)

    assert np.allclose(h[1].properties['Lx'].ratio("1e13 kpc Msol km s**-1"), -90787.3983020414)

@pytest.mark.parametrize("with_units", [True, False])
def test_swift_velociraptor_all_properties(with_units):
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    h = pynbody.halo.velociraptor.VelociraptorCatalogue(f)
    all_properties = h.get_properties_all_halos(with_units=with_units)

    assert np.allclose(all_properties['Lx'][0], -90787.3983020414)
    assert np.allclose(all_properties['Lx'][9], -73881.83861892551)

    if with_units:
        assert all_properties['Lx'].units == '1e13 kpc Msol km s**-1'
    else:
        assert not hasattr(all_properties['Lx'], 'units')

def test_swift_velociraptor_select_catalogue_with_filename():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    h = f.halos(filename="testdata/SWIFT/catalogue_0150/output")
    assert isinstance(h, pynbody.halo.velociraptor.VelociraptorCatalogue)
    h.load_all()
    assert len(h) == 209



import threading
import time
import weakref

import numpy as np
import pytest

import pynbody


def setup_module():
    global f, h
    f = pynbody.new(dm=1000, star=500, gas=500, order='gas,dm,star')
    f['pos'] = np.random.normal(scale=1.0, size=f['pos'].shape)
    f['vel'] = np.random.normal(scale=1.0, size=f['vel'].shape)
    f['mass'] = np.random.uniform(1.0, 10.0, size=f['mass'].shape)
    f.gas['rho'] = np.ones(500, dtype=float)


def teardown_module():
    global f
    del f


def test_array_promotion():
    f.dm['tx'] = np.arange(0, 1000)
    f.gas['tx'] = np.arange(0, 500)

    # currently still a family array
    assert 'tx' in f.family_keys()
    assert 'tx' not in f.keys()

    with pytest.raises(KeyError):
        f.star['tx']

    with pytest.raises(KeyError):
        f['tx']

    f.star['tx'] = np.arange(0, 500)

    # Should have been promoted
    assert 'tx' not in f.family_keys()
    assert 'tx' in f.keys()

    f['tx']  # should succeed
    del f['tx']


def test_one_family_promotion():
    fx = pynbody.new(dm=10)
    fx.dm['bla'] = np.arange(10)
    # should have been made as a full-simulation array

    assert 'bla' in fx.keys()
    fx['bla']
    del fx


def test_subscript():

    a = f[::37]
    assert(np.abs(a['pos'] - f['pos'][::37]).sum() < 1.e-10)

    a = f[[1, 5, 9, 12, 52, 94]]
    assert len(a) == 6


def test_noarray():
    def check(z, x):
        with pytest.raises(KeyError):
            z[x]

    check(f, 'thisarraydoesnotexist')
    check(f.gas, 'thisarraydoesnotexist')
    check(f[::5], 'thisarraydoesnotexist')
    check(f[[1, 2, 7, 93]], 'thisarraydoesnotexist')


def test_derived_array():
    assert 'vr' in f.derivable_keys()
    f['vr']
    assert 'vr' in f.keys()
    assert f['vr'].derived
    assert f.gas['vr'].derived
    assert f[::4]['vr'].derived
    assert f[[2, 3, 7, 12]]['vr'].derived
    f['pos'] += (2, 0, 0)
    assert 'vr' not in f.keys()
    f['pos'] -= (2, 0, 0)
    f.gas['vr']
    assert f.gas['vr'].derived
    f.dm['vr']
    assert f.dm['vr'].derived

    assert 'vr' not in f.keys()
    assert 'vr' in f.gas.keys()
    f['vr']
    assert 'vr' in f.keys()

    with pytest.raises(ValueError):
        f['r'][22] += 3

    f['r'].derived = False
    # Now we should be able to write to it
    f['r'][22] += 3

    f['pos'] += (2, 0, 0)
    # broken the link, so r shouldn't have been marked dirty
    assert 'r' in f.keys()


def test_unit_inheritance():
    f['pos'].units = 'km'
    f['vel'].units = 'km s^-1'


def test_equality():
    assert f == f
    assert f[[1, 5, 8]] != f
    assert f[[1, 6, 9]] != f[[1, 5, 22]]
    assert f[::5][[1, 2, 3]] == f[[5, 10, 15]]
    assert f.dm == f.dm
    assert f.dm == f[f._get_family_slice(pynbody.family.dm)]


def test_arraytype():
    SA = pynbody.array.SimArray
    ISA = pynbody.array.IndexedSimArray

    assert type(f["mass"]) is SA
    assert type(f[[1, 2, 3]]["mass"]) is ISA
    assert type(f[::20]["mass"]) is SA
    assert type(f.dm['mass']) is SA
    assert type(f.gas['rho']) is SA
    assert type(f[[1, 3, 5, 7, 9, 11]].dm['mass']) is ISA
    assert type(f[[1, 2, 3, 8]].gas['rho']) is ISA


def test_persistence():
    f.dm.kdtree = 123
    assert f.dm.kdtree == 123
    f.dm[::12].kdtree = 234
    assert f.dm[::12].kdtree == 234
    assert f.dm.kdtree == 123
    f.gas.kdtree = 96
    assert f.gas.kdtree == 96
    assert f.dm.kdtree == 123
    f.kdtree = 92
    assert f.kdtree == 92
    assert f.dm.kdtree == 123
    assert f.dm[::12].kdtree == 234
    assert f.gas.kdtree == 96


def test_copy():
    import copy
    f2 = copy.deepcopy(f)
    f2.dm['mass'][0] = 2
    f.dm['mass'][0] = 1
    assert f2.dm['mass'][0] != f.dm['mass'][0]

    assert len(f2) == len(f)
    assert len(f2.dm) == len(f.dm)
    assert len(f2.gas['rho']) == 500

    f2 = copy.deepcopy(f[::3])
    assert len(f2) == len(f[::3])
    assert len(f2.gas) == len(f[::3].gas)

    assert all(f2.gas['mass'] == f[::3].gas['mass'])
    f2.gas['mass'][0] = 999
    assert not all(f2.gas['mass'] == f[::3].gas['mass'])

    f['pos'].units = 'kpc'
    f2 = copy.deepcopy(f[[1, 2, 3, 4]])
    assert f2['pos'].units == 'kpc'
    # this tests for a bug where units were not correctly
    # copied in from IndexedSimArrays


def test_mean_by_mass_units():
    f['pos'].units = 'kpc'
    f['mass'].units = 'Msol'

    assert str(f.mean_by_mass('pos').units) == 'kpc'


def test_name_awareness():
    assert f['pos'].name == 'pos'


def test_immediate_mode():
    with f.immediate_mode:
        assert isinstance(f[[1, 6, 10]]['x'], pynbody.array.SimArray)
        test_val = f[[1, 6, 10]]['x']
    assert isinstance(f[[1, 6, 10]]['x'], pynbody.array.IndexedSimArray)
    assert (test_val == f[[1, 6, 10]]['x']).all()

    with f.immediate_mode:
        # check we get the same actual object two times
        fsub = f.dm[[1, 6, 52]]
        xa = fsub['x']
        xb = fsub['x']
        assert xa is xb
        xc = f.dm[[1, 6, 52]]['x']
        assert xa is xc
        wr = weakref.ref(xa)
        del xa, xb, xc
        assert wr() is not None

    # check it was deleted
    assert wr() is None

    del f['r']

    f.dm['r']

    # check this also works with family-level arrays

    with f.immediate_mode:
        assert isinstance(f.dm[[22, 53, 69]]['r'], pynbody.array.SimArray)
        assert isinstance(f[[600, 603, 670]].dm['r'], pynbody.array.SimArray)


def test_subsnap_by_boolean_mask():
    print((f['x'] > 0).shape, len(f))
    assert (f['x'][f['x'] > 0] == f[f['x'] > 0]['x']).all()

def test_issue_206() :
    assert len(f.s[[1,4,29]].s)==3


def test_nd_array_slicing():
    assert np.all(f['x']==f['pos'][:,0])
    assert f._array_name_1D_to_ND('x')=='pos'
    assert f._array_name_1D_to_ND('anything_y')=='anything'

    pos_components = set(f._array_name_ND_to_1D('pos'))
    assert pos_components=={'x','y','z'}

    anything_components = set(f._array_name_ND_to_1D('anything'))
    assert anything_components=={'anything_x','anything_y','anything_z'}

    assert f._array_name_implies_ND_slice('x')

    # following is only true if the array 'anything' actually exists in loadable_keys, which it doesn't
    assert not f._array_name_implies_ND_slice('anything_x')

    # the judgement is on name only, so even though mass_x could not be created it is judged as
    # a 1D slice of the 'mass' array:
    assert f._array_name_implies_ND_slice('mass_x')
    assert f.gas._array_name_implies_ND_slice('rho_x')

    # rho doesn't exist in f.loadable_keys(), only f.gas.loadable_keys():
    assert not f._array_name_implies_ND_slice('rho_x')

def test_index_list():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    h = f.halos()
    index_list = h[1].get_index_list(f)

def test_mean_by_mass_value():
    f = pynbody.new(dm=2000)
    f['mass'] = np.ones(2000)
    f['mass'][:500] = 2

    f['test_array'] = np.arange(2000)

    f = f[np.arange(1000)] # test on a subsnap for generality

    np.testing.assert_allclose(f.mean_by_mass('test_array'),
                               (f['test_array'][:500]*2./3 + f['test_array'][500:1000]*1./3).mean())


class SlowLoadingSnapshot(pynbody.snapshot.simsnap.SimSnap):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._has_entered = False
        self._has_exited = False
        self._filename = "test"

    def _load_array(self, array_name, fam=None):
        if self._has_entered:
            raise RuntimeError("Entered twice")
        if self._has_exited:
            raise RuntimeError("Entered after exiting")
        self._has_entered=True
        time.sleep(0.1)
        self[array_name] = np.arange(len(self))
        self._has_entered=False
        self._has_exited=True

def test_race_condition():
    """There was a race condition where if multiple threads tried to load the same array at once, strange errors
    would result. This has been fixed by a lock on deriving/loading data."""
    f = SlowLoadingSnapshot()

    def load_array():
        f['anyarray'] # noqa

    threads = [threading.Thread(target=load_array) for i in range(5)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()



"""Read and expose configuration information for pynbody

The configuration system in pynbody is described in the :ref:`configuration` tutorial.

"""

import logging
import multiprocessing
import os
import warnings


def _get_config_parser_with_defaults():
    # Create config dictionaries which will be required by subpackages
    import configparser
    config_parser = configparser.ConfigParser()
    config_parser.optionxform = str
    config_parser.read(
        os.path.join(os.path.dirname(__file__), "default_config.ini"))
    return config_parser

def _merge_defaults_for_problematic_keys(config_parser):
    """This unfortunate routine is made necessary by issue #261"""
    config_parser_defaults = _get_config_parser_with_defaults()
    merge = (('irreducible-units','names'),)

    for merge_i in merge:
        opt = config_parser.get(*merge_i)
        default_opt = config_parser_defaults.get(*merge_i)

        items = list(map(str.strip,opt.split(",")))
        default_items = list(map(str.strip,default_opt.split(",")))

        for checking_item in default_items:
            if checking_item not in items:
                warnings.warn("Pynbody spotted a potential problem with your .pynbodyrc or config.ini. Overriding it by adding %r to config section %r item %r"%(checking_item, merge_i[0], merge_i[1]))
                opt+=", "+checking_item
                config_parser.set(merge_i[0],merge_i[1],opt)

def _add_overrides_to_config_parser(config_parser):
    config_parser.read(os.path.join(os.path.dirname(__file__), "config.ini"))
    config_parser.read(os.path.expanduser("~/.pynbodyrc"))
    config_parser.read("config.ini")
    _merge_defaults_for_problematic_keys(config_parser)

def _get_basic_config_from_parser(config_parser):

    config = {'centering-scheme': config_parser.get('general', 'centering-scheme')}

    config['snap-class-priority'] = list(map(str.strip,
                                        config_parser.get('general', 'snap-class-priority').split(",")))
    config['halo-class-priority'] = list(map(str.strip,
                                        config_parser.get('general', 'halo-class-priority').split(",")))


    config['default-cosmology'] = {}
    for k in config_parser.options('default-cosmology'):
        config[
            'default-cosmology'][k] = float(config_parser.get('default-cosmology', k))

    config['sph'] = {}
    for k in config_parser.options('sph'):
        try:
            config['sph'][k] = int(config_parser.get('sph', k))
        except ValueError:
            pass
    config['sph']['Kernel'] = config_parser.get('sph', 'Kernel')

    config['threading'] = config_parser.get('general', 'threading')
    config['number_of_threads'] = int(
        config_parser.get('general', 'number_of_threads'))

    if config['number_of_threads']<0:
        config['number_of_threads']=multiprocessing.cpu_count()

    config['gravity_calculation_mode'] = config_parser.get(
        'general', 'gravity_calculation_mode')
    config['disk-fit-function'] = config_parser.get('general', 'disk-fit-function')

    config['image-default-resolution'] = int(config_parser.get('general', 'image-default-resolution'))

    return config

def _setup_logger(config):
    logger = logging.getLogger('pynbody')
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s : %(message)s')
    for existing_handler in list(logger.handlers):
        logger.removeHandler(existing_handler)

    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if config_parser.getboolean('general','verbose'):
        set_logging_level(logging.INFO)
        logger.info("Verbose mode is on")
    else:
        set_logging_level(logging.WARNING)

def set_logging_level(level = logging.INFO):
    """Set the logging level for pynbody, in terms of the standard Python logging module levels.

    Set to logging.INFO for more verbose output, or logging.WARNING for less."""
    logger = logging.getLogger('pynbody')
    logger.setLevel(level)


config_parser = _get_config_parser_with_defaults()
_add_overrides_to_config_parser(config_parser)
config = _get_basic_config_from_parser(config_parser)
logger = _setup_logger(config)



"""Support for numpy arrays in shared memory."""
import atexit
import functools
import mmap
import os
import random
import signal
import time
from functools import reduce

import numpy as np
import posix_ipc

from ..configuration import config_parser
from . import SimArray

_owned_shared_memory_names = []

class SharedArrayNotFound(OSError):
    pass

class SharedMemorySimArray(SimArray):
    __slots__ = ['_shared_fname', '_shared_owner']
    _shared_fname: str
    _shared_owner: bool

    def __del__(self):
        global _owned_shared_memory_names
        if hasattr(self, '_shared_fname') and getattr(self, '_shared_owner', False):
            if self._shared_fname in _owned_shared_memory_names:
                _owned_shared_memory_names.remove(self._shared_fname)
                posix_ipc.unlink_shared_memory(self._shared_fname)

def make_shared_array(dims, dtype, zeros=False, fname=None, create=True,
                      offset = None, strides = None) -> SharedMemorySimArray:
    """Create an array of dimensions *dims* with the given numpy *dtype*.

    Parameters
    ----------

    dims: tuple | int
       The dimensions of the numpy array to create
    dtype:
       The data type to use
    zeros:
        If True, zero the array; otherwise leave uninitialised
    fname: str, optional
        The shared memory name to use. If None, a random name will be created. This is only
        valid with create=True.
    create: bool
        Whether to create the shared array, or to open existing shared memory. If the latter,
        the fname must be specified and the caller is responsible for making sure the dtype
        and dims match the original array.
    offset: int, optional
        The offset into the shared memory to use. This is only valid with create=False
    strides: tuple, optional
        The strides to use. This is only valid with create=False
    """
    if fname is None:
        if not create:
            raise ValueError("When opening an existing shared array, fname must be specified")
        random.seed(os.getpid() * time.time())
        fname = "pynbody-" + \
                ("".join([random.choice('abcdefghijklmnopqrstuvwxyz')
                          for _ in range(10)]))

    if not hasattr(dims, '__len__'):
        dims = (dims,)

    zero_size = False
    if dims[0] == 0:
        zero_size = True
        dims = (1,) + dims[1:]
    if hasattr(dims, '__len__'):
        size = reduce(np.multiply, dims)
    else:
        size = dims

    if create:
        _register_sigterm_handler()
        if offset is not None:
            raise ValueError("Offset only valid when opening an existing shared array")
        if strides is not None:
            raise ValueError("Strides only valid when opening an existing shared array")
        mem = posix_ipc.SharedMemory(fname, posix_ipc.O_CREX, size=int(np.dtype(dtype).itemsize*size))
        _owned_shared_memory_names.append(fname)

    else:
        try:
            mem = posix_ipc.SharedMemory(fname)
        except posix_ipc.ExistentialError:
            raise SharedArrayNotFound(f"No shared memory found with name {fname}") from None

    mapped_mem = mmap.mmap(mem.fd, mem.size)
    mem.close_fd()

    if offset is None:
        offset = 0

    ret_ar = np.frombuffer(mapped_mem, dtype=dtype, count=size, offset=offset).reshape(dims).view(SharedMemorySimArray)
    ret_ar._shared_fname = fname
    ret_ar._shared_owner = create

    if strides:
        ret_ar.strides = strides

    if zero_size:
        ret_ar = ret_ar[1:]

    if zeros:
        ret_ar[:] = 0

    return ret_ar

@atexit.register
def delete_dangling_shared_memory():
    """Ensures that all shared memory has been cleaned up."""
    global _owned_shared_memory_names

    for fname in _owned_shared_memory_names:
        try:
            posix_ipc.unlink_shared_memory(fname)
        except posix_ipc.ExistentialError:
            pass
    _owned_shared_memory_names = []

_sigterm_handler_is_registered = False

def _sigterm_handler(signum, frame):
    delete_dangling_shared_memory()

def _register_sigterm_handler():
    """Registers a handler to clean up shared memory in the event of a SIGTERM signal."""
    global _sigterm_handler_is_registered
    if not _sigterm_handler_is_registered:
        signal.signal(signal.SIGTERM, _sigterm_handler)
        _sigterm_handler_is_registered = True

def get_num_shared_arrays_owned():
    """Returns the number of shared arrays currently owned by this process.

    A shared array is only considered owned if this process is reponsible for unlinking it on exit.
    """
    return len(_owned_shared_memory_names)

class _deconstructed_shared_array(tuple):
    pass

def _shared_array_deconstruct(ar, transfer_ownership=False):
    """Deconstruct an array backed onto shared memory into something that can be
    passed between processes efficiently. If *transfer_ownership* is True,
    also transfers responsibility for deleting the underlying memory (if this
    process has it) to the reconstructing process."""

    assert isinstance(ar, SimArray)
    ar_base = ar
    while isinstance(ar_base.base, SimArray):
        ar_base = ar_base.base

    assert isinstance(ar_base, SharedMemorySimArray), "Cannot prepare an array for shared use unless it was created in shared memory"

    ownership_out = transfer_ownership and ar_base._shared_del
    if transfer_ownership:
        ar_base._shared_del = False

    offset = ar.__array_interface__['data'][0] - \
              ar_base.__array_interface__['data'][0]

    return _deconstructed_shared_array((ar.dtype, ar.shape, ar_base._shared_fname, ownership_out,
                                        offset, ar.strides))


def _shared_array_reconstruct(X):
    dtype, dims, fname, ownership, offset, strides = X

    assert not ownership # transferring ownership not actually supported in current implementation

    new_ar = make_shared_array(dims, dtype, fname=fname, create=False, offset=offset, strides=strides)

    return new_ar


def _recursive_shared_array_deconstruct(input, transfer_ownership=False) :
    """Works through items in input, deconstructing any shared memory arrays
    into transferrable references"""
    output = []
    for item in input:
        if isinstance(item, SimArray):
            item = _shared_array_deconstruct(item, transfer_ownership)
        elif isinstance(item, list) or isinstance(item, tuple):
            item = _recursive_shared_array_deconstruct(item, transfer_ownership)
        output.append(item)
    return output


def _recursive_shared_array_reconstruct(input):
    """Works through items in input, reconstructing any shared memory arrays
    from transferrable references"""
    output = []
    for item in input:
        if isinstance(item, _deconstructed_shared_array):
            item = _shared_array_reconstruct(item)
        elif isinstance(item, list) or isinstance(item, tuple):
            item = _recursive_shared_array_reconstruct(item)
        output.append(item)
    return output


class RemoteKeyboardInterrupt(Exception):
    pass


def shared_array_remote(fn):
    """A decorator for functions that are expected to run on a 'remote' process, accepting shared memory inputs.

    The decorator reconstructs any shared memory arrays that have been 'deconstructed' into a reference by remote_map
    """

    @functools.wraps(fn)
    def new_fn(args, **kwargs):
        try:
            import signal
            assert hasattr(
                args, '__len__'), "Function must be called from remote_map to use shared arrays"
            assert args[0] == '__pynbody_remote_array__', "Function must be called from remote_map to use shared arrays"
            args = _recursive_shared_array_reconstruct(args)
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            output = fn(*args[1:], **kwargs)
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            return _recursive_shared_array_deconstruct([output], True)[0]
        except KeyboardInterrupt:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            raise RemoteKeyboardInterrupt()
    new_fn.__pynbody_remote_array__ = True

    return new_fn


def remote_map(pool, fn, *iterables):
    """Equivalent to pool.map, but turns any shared memory arrays into a reference that can be passed between processes.

    The function *fn* must be wrapped with the *shared_array_remote* decorator for this to work correctly."""

    assert getattr(fn, '__pynbody_remote_array__',
                   False), "Function must be wrapped with shared_array_remote to use shared arrays"
    iterables_deconstructed = _recursive_shared_array_deconstruct(
        iterables)
    try:
        results = pool.map(fn, list(zip(
            ['__pynbody_remote_array__'] * len(iterables_deconstructed[0]), *iterables_deconstructed)))
    except RemoteKeyboardInterrupt:
        raise KeyboardInterrupt
    return _recursive_shared_array_reconstruct(results)



"""Tools for tracking how derived arrays depend on other arrays

Users should not need to use this module directly; it is used by :class:`~pynbody.snapshot.simsnap.SimSnap`
to track dependencies between derived arrays.
"""

import threading


class DependencyError(RuntimeError):
    pass

class _DependencyContext:
    """Context manager for tracking dependencies between arrays"""
    def __init__(self, tracker, name):
        self.tracker = tracker
        self.name = name

    def __enter__(self):
        self.tracker._calculation_lock.__enter__()
        if self.name in self.tracker._current_calculation_stack:
            self.tracker._calculation_lock.__exit__(None, None, None)
            raise DependencyError("Circular dependency")
        self.tracker._push(self.name)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tracker._pop(self.name)
        if exc_type is None:
            self.tracker._add_me_to_dependents(self.name)
        self.tracker._calculation_lock.__exit__(exc_type, exc_val, exc_tb)


class DependencyTracker:
    """Class for tracking dependencies between arrays.

    This class is used by :class:`pynbody.snapshot.simsnap.SimSnap` to track how derived arrays
    depend on other arrays. Users should not need to use this class directly.

    As an example of how this class is used, see the following code.

    >>> my_tracker = DependencyTracker()
    >>> with my_tracker.calculating('my_array'):
    >>>    with my_tracker.calculating('my_other_array'):
    >>>       my_tracker.touching('source_array')
    >>> my_tracker.get_dependents('my_array') # -> {}
    >>> my_tracker.get_dependents('my_other_array') # -> {'my_array'}
    >>> my_tracker.get_dependents('source_array') # -> {'my_other_array', 'my_array'}
    >>> with my_tracker.calculating('my_array'):
    >>>     with my_tracker.calculating('my_array'): # -> raises DependencyError
    >>>        pass

    Note that the class is thread-safe, in the sense that if a second thread starts trying to use it
    while a first thread is already mid-way through a derivation, the second thread blocks until
    the first thread finishes.
    """
    def __init__(self):
        self._dependencies = {}
        self._current_calculation_stack = []
        self._calculation_lock = threading.RLock()

    def _setup_my_dependencies(self,name):
        if name not in self._dependencies:
            self._dependencies[name] = set()

    def _add_me_to_dependents(self, name):
        self._setup_my_dependencies(name)
        for other in self._current_calculation_stack:
            self._dependencies[name].add(other)

    def _push(self, name):
        self._current_calculation_stack.append(name)

    def _pop(self, name):
        assert self._current_calculation_stack[-1]==name
        del self._current_calculation_stack[-1]


    def calculating(self, name):
        """Return a context manager when calculating a named array."""
        return _DependencyContext(self, name)

    def touching(self,name):
        """Note that any ongoing calculations depend on the named array."""
        with self._calculation_lock:
            self._add_me_to_dependents(name)

    def get_dependents(self, name):
        """Return the set of arrays that are known to depend on the named array."""
        return self._dependencies.get(name, set())



import os.path
import re
import struct

import numpy as np

from .number_array import HaloNumberCatalogue


class HOPCatalogue(HaloNumberCatalogue):
    """Represents a HOP Catalogue as used by Ramses."""

    dtype = np.int32

    def __init__(self, sim, filename=None):
        """Create a new HOP catalogue object from the given simulation object.

        Parameters
        ----------

        sim : :class:`pynbody.snapshot.SimSnap`
            The simulation snapshot to which this catalogue applies.

        filename : str, optional
            The filename of the HOP catalogue. The filename should be either the ``grp*.tag`` file or the
            ``hop*.hop`` file that HOP produces. If no filename is provided, the code will attempt to find a
            suitable ``grp*.tag`` file in the Ramses output folder, in the folder in which the simulation is
            located, or in a ``hop`` subfolder.

        """

        if filename is None:
            for filename in HOPCatalogue._enumerate_hop_tag_locations_from_sim(sim):
                if os.path.exists(filename):
                    break

            if not os.path.exists(filename):
                raise RuntimeError("Unable to find HOP .tag file in simulation directory")

        sim._create_array('hop_grp', dtype=self.dtype)
        sim['hop_grp'] = -1
        with open(filename, "rb") as f:
            num_part = self._get_npart_from_file(f)

            if num_part != len(sim.dm):
                raise RuntimeError("Mismatching number of particles between "
                                   "snapshot %s and HOP file %s. Check your pynbody configuration for any missing"
                                   " particle fields or partial loading" % (sim.filename, filename))

            sim.dm['hop_grp'] = np.fromfile(f, self.dtype, len(sim.dm))
        super().__init__(sim, array="hop_grp", ignore=-1)

    @classmethod
    def _get_npart_from_file(cls, f):
        num_part, = struct.unpack('i', f.read(4))
        if num_part == 8:
            # fortran-formatted output
            num_part, num_grps, _, _ = struct.unpack('iiii', f.read(16))
        else:
            # plain binary output
            num_grps, = struct.unpack('i', f.read(4))
        return num_part

    @classmethod
    def _can_load(cls, sim, filename=None):
        # Hop output must be in output directory or in output_*/hop directory
        if filename is not None:
            if not os.path.exists(filename):
                return False
            with open(filename, "rb") as f:
                num_part = cls._get_npart_from_file(f)
            return num_part == len(sim.dm)
        else:
            exists = any([os.path.exists(fname) for fname in HOPCatalogue._enumerate_hop_tag_locations_from_sim(sim)])
            return exists

    @staticmethod
    def _extract_hop_name_from_sim(sim):
        match = re.search("output_([0-9]*)", sim.filename)
        if match is None:
            raise OSError("Cannot guess the HOP catalogue filename for %s" % sim.filename)
        return "grp%s.tag" % match.group(1)

    @staticmethod
    def _enumerate_hop_tag_locations_from_sim(sim):
        try:
            name = HOPCatalogue._extract_hop_name_from_sim(sim)
        except OSError:
            return []

        s_filename = os.path.abspath(sim.filename)

        return [os.path.join(os.path.dirname(s_filename),name),
                os.path.join(s_filename,name),
                os.path.join(s_filename,'hop',name)]



"""
Implements loading of nchilada files.

"""

import os
import struct
import warnings
import xml.dom.minidom

import numpy as np

from .. import chunk, family, units
from . import SimSnap, namemapper

_name_map, _rev_name_map = namemapper.setup_name_maps('nchilada-name-mapping')
_translate_array_name = namemapper.name_map_function(_name_map, _rev_name_map)

_type_codes = list(map(np.dtype, [None, 'int8', 'uint8', 'int16', 'uint16',
                             'int32', 'uint32', 'int64', 'uint64',
                             'float32', 'float64']))


_max_buf = 1024 * 512


class NchiladaSnap(SimSnap):
    """Implements loading of nchilada files."""

    def _load_header(self, f):
        # Used to use xdrlib, but that is deprecated in Python 3.11
        # The following code just follows the old xdrlib implementation
        buf= f.read(28)
        pos= 0
        # Read int, walrus assignment updates pos
        assert struct.unpack('>l',buf[pos:(pos:=pos+4)])[0] == 1062053
        # Read double, walrus assignment updates pos
        t= struct.unpack('>d',buf[pos:(pos:=pos+8)])[0]
        # Read four ints, walrus assignment updates pos
        highword, nbod, ndim, code = (
            struct.unpack('>l',buf[pos:(pos:=pos+4)])[0] for i in range(4)
        )
        return t, nbod, ndim, _type_codes[code]

    def _update_loadable_keys(self):
        self._loadable_keys_registry = {}
        d = self._loadable_keys_registry
        fams = self._dom_sim.getElementsByTagName('family')
        for f in sorted(fams, key=lambda x: str(x)):
            fam_name = f.attributes['name'].value
            our_fam = family.get_family(fam_name)
            d_f = {}
            for a in f.getElementsByTagName('attribute'):
                our_name = _translate_array_name(
                    a.attributes['name'].value, reverse=True)
                filename = os.path.join(
                    self._filename, a.attributes['link'].value)
                d_f[our_name] = filename
            d[our_fam] = d_f

    def _setup_slices(self, take=None):
        disk_family_slice = {}
        i = 0
        # for each family, find an array (any array) to determine length and
        # set up a logical map of particles on disk
        for f in sorted(self._loadable_keys_registry.keys()):
            ars = self._loadable_keys_registry[f]
            with open(list(ars.values())[0], 'rb') as tf:
                header_time, nbod, _, _ = self._load_header(tf)
            disk_family_slice[f] = slice(i, i + nbod)
            i += nbod
            self.properties['a'] = header_time
        self._load_control = chunk.LoadControl(
            disk_family_slice, _max_buf, take)
        self._family_slice = self._load_control.mem_family_slice
        self._num_particles = self._load_control.mem_num_particles

    def __init__(self, filename, **kwargs):
        """Load a nchilada file.

        Parameters
        ----------

        filename : str
            The path to the nchilada output to load. Nchilada outputs are directories containing a description.xml
            file and a number of binary files.
        take : np.ndarray, optional
            The array of particles to load. If not specified, all particles are loaded.
        """

        super().__init__()

        must_have_paramfile = kwargs.get('must_have_paramfile', False)
        take = kwargs.get('take', None)

        self._dom_sim = xml.dom.minidom.parse(
            os.path.join(filename, "description.xml")).getElementsByTagName('simulation')[0]
        self._filename = filename
        self._update_loadable_keys()
        self._setup_slices(take=take)
        self._paramfilename = kwargs.get('paramfile', None)

        self._decorate()

        if 'dKpcUnit' not in self._paramfile:
            if must_have_paramfile:
                raise RuntimeError("Could not find .param file for this run. Place it in the run's directory or parent directory.")
            else:
                warnings.warn(
                    "No readable param file in the run directory or parent directory: using defaults.", RuntimeWarning)
                self._file_units_system = [
                    units.Unit(x) for x in ('G', '1 kpc', '1e10 Msol')]


    def loadable_keys(self, fam=None):
        if fam is not None:
            return list(self._loadable_keys_registry[fam].keys())
        else:
            loadable = None
            for f in self._loadable_keys_registry.values():
                if loadable is None:
                    loadable = set(f.keys())
                else:
                    loadable = loadable.intersection(iter(f.keys()))
            return list(loadable)

    def _open_file_for_array(self, fam, array_name):
        fname = self._loadable_keys_registry[fam].get(array_name, None)
        if not fname:
            raise OSError("No such array on disk")
        f = open(fname, 'rb')
        return f

    def _attempt_load_all_families(self, array_name):
        fams = self.families()
        universal_dtype = None
        universal_ndim = None
        if fams==[]:
            return
        for fam in fams:
            # this will raise an IOError propagating upwards if any family doesn't have the appropriate array
            with self._open_file_for_array(fam, array_name) as f:
                _, nbod, ndim, dtype = self._load_header(f)
            if universal_dtype is not None:
                if ndim!=universal_ndim:
                    raise OSError("Mismatching dimensions for array")
                if dtype!=universal_dtype:
                    raise OSError("Mismatching data type for array")
            universal_ndim, universal_dtype = ndim, dtype

        self._create_array(array_name,universal_ndim,universal_dtype,False)

        for fam in fams:
            self._load_array(array_name, fam)


    def _load_array(self, array_name, fam=None):
        if fam is None:
            self._attempt_load_all_families(array_name)
            return

        f = self._open_file_for_array(fam, array_name)

        _, nbod, ndim, dtype = self._load_header(f)
        if array_name not in list(self.keys()):
            self._create_family_array(array_name, fam, ndim=ndim,dtype=dtype)
        r = self[fam][array_name]
        if units.has_units(r):
            r.convert_units(self._default_units_for(array_name))
        else:
            r.set_default_units()

        disk_dtype = dtype.newbyteorder('>')

        buf_shape = _max_buf
        if ndim > 1:
            buf_shape = (_max_buf, ndim)

        b = np.empty(buf_shape)

        # skip over min and max values (see issue #211)
        np.fromfile(f, dtype=disk_dtype, count=2 * ndim)

        for readlen, buf_index, mem_index in self._load_control.iterate(fam, fam):
            b = np.fromfile(f, dtype=disk_dtype, count=readlen * ndim)
            if ndim > 1:
                b = b.reshape((readlen, ndim))

            if mem_index is not None:
                r[mem_index] = b[buf_index]
        f.close()

    @classmethod
    def _can_load(cls, f):
        return os.path.isdir(f) and os.path.exists(os.path.join(f, "description.xml"))



"""
Support for halo and group catalogues.

Halo catalogues act like a dictionary, mapping from halo numbers to a Halo objects. The halo *number* is typically
determined by the halo finder, and is often (but not always) the same as the halo *index* which is the zero-based
offset within the catalogue.

If you have a supported halo catalogue on disk or a halo finder installed and correctly configured, you can access a
halo catalogue through ``f.halos()`` where ``f`` is a SimSnap.

See the :ref:`halo catalogue tutorial <halo_tutorial>` for introductory information and guidance.


.. _v2_0_halo_changes:

.. versionchanged:: 2.0

  Backwards-incompatible changes to the halo catalogue system

  For version 2.0, the halo catalogue loading system was substantially rewritten. The new system is more robust and
  more consistent across different halo finders. However, this means that some defaults have changed, most significantly
  in the AHF halo numbering. Backward-compatibility can be achieved by passing ``halo_numbers='v1'`` to the
  :class:`~pynbody.halo.ahf.AHFCatalogue` constructor. For more information, read the documentation for that class.

  Furthermore, older versions of pynbody (i.e. v1.x) could be configured to create a halo catalogue if one was not
  found, using AHF. This is no longer the case, as creating a halo catalogue requires choosing a halo finder and its
  parameters carefully for the task in hand and it was not possible to provide a one-size-fits-all solution.

  Finally, options to write ``.stat`` files and ``.grp`` files have been removed. However it is still possible to
  generate a ``.grp`` file by  calling :meth:`~HaloCatalogue.get_group_array` and writing out the resulting
  array of integers using a tool like ``numpy.savetxt``.

  By paring back the less-used functionality of the halo catalogue system, the remaining functionality is more
  consistent, robust, and extensible to new halo finders.


.. _supported_halo_finders:

Supported halo-finder formats
-----------------------------

The currently-supported formats are:

- Adaptahop (:class:`~pynbody.halo.adaptahop.AdaptaHOPCatalogue`);
- AHF (:class:`~pynbody.halo.ahf.AHFCatalogue`);
- HBT+ (:class:`~pynbody.halo.hbtplus.HBTPlusCatalogue`);
- HOP (:class:`~pynbody.halo.hop.HOPCatalogue`);
- Rockstar (:class:`~pynbody.halo.rockstar.RockstarCatalogue`);
- Subfind (old format :class:`~pynbody.halo.subfind.SubfindCatalogue`, or various HDF5 variants
  as :class:`~pynbody.halo.subfindhdf.SubfindHDFCatalogue`);
- VELOCIraptor (:class:`~pynbody.halo.velociraptor.VelociraptorCatalogue`).

In addition, generic halo finders which output a list of halo numbers for each particle are supported via
:class:`~pynbody.halo.number_array.HaloNumberCatalogue`.



.. note::

    The principal development of ``pynbody`` took place in the UK, and the spelling of "catalogue" is British English.
    However, since much code is written in American English, v2.0.0 introduced aliases such that all
    classes can be accessed with the American spelling ``HaloCatalog``, ``AdaptaHOPCatalog`` etc.


"""
from __future__ import annotations

import copy
import logging
import warnings
import weakref
from typing import TYPE_CHECKING, Iterable

import numpy as np
from numpy.typing import NDArray

from .. import array, snapshot, units, util
from ..util import iter_subclasses
from .details.iord_mapping import make_iord_to_offset_mapper
from .details.number_mapping import MonotonicHaloNumberMapper, create_halo_number_mapper
from .details.particle_indices import HaloParticleIndices

if TYPE_CHECKING:
    from .subhalo_catalogue import SubhaloCatalogue

logger = logging.getLogger("pynbody.halo")

class DummyHalo(snapshot.util.ContainerWithPhysicalUnitsOption):

    def __init__(self):
        self.properties = {}

    def physical_units(self, *args, **kwargs):
        pass


class Halo(snapshot.subsnap.IndexedSubSnap):

    """
    Represents a single halo from a halo catalogue.

    Note that pynbody refers to groups, halos and subhalos interchangably, with the term "halo" being used to cover
    all of these.
    """

    def __init__(self, halo_number, properties, halo_catalogue, *args, **kwa):
        super().__init__(*args, **kwa)
        self._halo_catalogue = halo_catalogue
        self._halo_number = halo_number
        self._descriptor = "halo_" + str(halo_number)
        self.properties = copy.copy(self.properties)
        self.properties['halo_number'] = halo_number
        self.properties.update(properties)

        # Inherit autoconversion from parent
        self._autoconvert_properties()

    @property
    @util.deprecated("The sub property has been renamed to subhalos")
    def sub(self):
        """Deprecated alias for :property:`subhalos`."""
        return self.subhalos

    @property
    def subhalos(self) -> SubhaloCatalogue:
        """A HaloCatalogue object containing only the subhalos of this halo."""
        return self._halo_catalogue._get_subhalo_catalogue(self._halo_number)


    def physical_units(self, distance='kpc', velocity='km s^-1', mass='Msol', persistent=True, convert_parent=True):
        if convert_parent:
            self._halo_catalogue.physical_units(
                distance=distance,
                velocity=velocity,
                mass=mass,
                persistent=persistent
            )
        else:
            # Convert own properties
            self._autoconvert_properties()


class HaloCatalogue(snapshot.util.ContainerWithPhysicalUnitsOption,
                    iter_subclasses.IterableSubclasses):

    """Generic halo catalogue object.

    To the user, this presents a simple interface where calling ``h[i]`` returns halo ``i``. Properties of halos
    can be retrieved without loading the halo via :meth:`get_properties_one_halo` or :meth:`get_properties_all_halos`.

    More information for users can be found in the :ref:`halo catalogue tutorial <halo_tutorial>`; see also the
    :ref:`supported halo finders <supported_halo_finders>`.

    Implementing a new format
    ^^^^^^^^^^^^^^^^^^^^^^^^^

    To support a new format, subclass :class:`HaloCatalogue` and implement the following methods:

    * :meth:`__init__`
    * :meth:`_can_load`
    * :meth:`_get_all_particle_indices`
    * :meth:`_get_particle_indices_one_halo` [only if it's possible to do this more efficiently than
      :meth:`_get_all_particle_indices` for users accessing only a few halos]
    * :meth:`get_properties_all_halos` [only if you have halo finder-provided properties to expose]
    * :meth:`get_properties_one_halo` [only if you have halo finder-provided properties to expose, and it's efficient
      to expose them one halo at a time; the default implementation will call get_properties_all_halos and extract]
    * :meth:`get_group_array` [only if it's possible to do this more efficiently than the default implementation]

    Nomenclature/conventions are worth being aware of if you are implementing a new format:

    * The halo number is the user-exposed identifier for a halo. It is typically assigned by the halo finder, although
      subclasses are free to assign their own (e.g. some have a `halo_number` option that can be passed to the
      constructor to override the halo finder's numbering). The halo numbers are used to access individual halos via
      the [] operator.
    * The halo *index* is the zero-based offset within the catalogue, which may be different from the halo number.
      Internally, *pynbody* converts between these using a :class:`details.number_mapping.HaloNumberMapper` object,
      which is set up in the :meth:`__init__` method.
    * Particle indices should be returned from methods like :meth:`_get_particle_indices_one_halo` as zero-relative
      offsets within the snapshot, not particle IDs or 'iord's. Many halo finders output particle IDs which must
      therefore be mapped. To aid this, call :meth:`_init_iord_to_fpos` in your :meth:`__init__` method, which creates
      a mapper as :attr:`_iord_to_fpos`. See :mod:`details.iord_mapping` for more information.

    """

    def __init__(self, sim, number_mapper):
        self._base: weakref[snapshot.SimSnap] = weakref.ref(sim)
        self.number_mapper: HaloNumberMapper = number_mapper
        self._index_lists: HaloParticleIndices | None = None
        self._properties: dict | None = None
        self._cached_halos: dict[int, Halo] = {}
        self._persistent_units = None

    def load_all(self):
        """Loads all halos, which is normally more efficient if a large fraction of them will be accessed."""
        if not self._index_lists:
            index_lists = self._get_all_particle_indices()
            properties = self.get_properties_all_halos(with_units=True)
            if isinstance(index_lists, tuple):
                index_lists = HaloParticleIndices(*index_lists)
            self._index_lists = index_lists
            if len(properties)>0:
                self._properties = properties

            if self._persistent_units is not None:
                self._cached_properties_to_physical_units(self._persistent_units)

    @util.deprecated("precalculate has been renamed to load_all")
    def precalculate(self):
        """Deprecated alias for :meth:`load_all`"""
        self.load_all()

    def _get_num_halos(self):
        return len(self.number_mapper)

    def _get_all_particle_indices_cached(self):
        """Get the index information for all halos, using a cached version if available"""
        self.load_all()
        return self._index_lists

    def _get_all_particle_indices(self) -> HaloParticleIndices | tuple[np.ndarray, np.ndarray]:
        """Returns information about the index list for all halos.

        Returns an HaloParticleIndices object, which is a container for the following information:
        - particle_ids: particle IDs contained in halos, sorted by halo ID
        - boundaries: the indices in particle_ids where each halo starts and ends
        """
        raise NotImplementedError("This halo catalogue does not support loading all halos at once")

    def get_properties_one_halo(self, halo_number) -> dict:
        """Returns a dictionary of properties for a single halo, given a halo_number """

        # Default implementation: extract from all halos. Subclasses may override this if they can load properties
        # for a single halo more efficiently.
        self._properties = self.get_properties_all_halos(with_units=True)
        return self._get_properties_one_halo_using_cache_if_available(halo_number,
                                                                      self.number_mapper.number_to_index(halo_number))

    def get_properties_all_halos(self, with_units=True) -> dict:
        """Returns a dictionary of properties for all halos.

        If with_units is True, the properties are returned as SimArrays with units if possible. Otherwise, numpy arrays
        are returned.

        Note that the returned properties are in contiguous arrays, and as a result may be in a different order to the
        halo numbers which are used to access individual halos. To map between halo numbers and properties, use the
        .number_mapper object; or access individual property dictionaries by halo number using get_properties_one_halo."""
        return {}

    def _get_properties_one_halo_using_cache_if_available(self, halo_number, halo_index):
        if self._properties is None:
            return self.get_properties_one_halo(halo_number)
        else:
            return {k: units.get_item_with_unit(self._properties[k],halo_index)
                    for k in self._properties}

    def _get_particle_indices_one_halo(self, halo_number) -> NDArray[int]:
        """Get the index list for a single halo, given a halo_number.

        A generic implementation is provided that fetches index lists for all halos and then extracts the one"""
        self.load_all()
        return self._index_lists.get_particle_index_list_for_halo(
            self.number_mapper.number_to_index(halo_number)
        )

    def _get_particle_indices_one_halo_using_list_if_available(self, halo_number, halo_index) -> NDArray[int]:
        if self._index_lists:
            return self._index_lists.get_particle_index_list_for_halo(halo_index)
        else:
            if len(self._cached_halos) == 5:
                warnings.warn("Accessing multiple halos may be more efficient if you call load_all() on the "
                              "halo catalogue", RuntimeWarning)
            return self._get_particle_indices_one_halo(halo_number)
            # NB subclasses may implement loading one halo direct from disk in the above
            # if not, the default implementation will populate _cached_index_lists

    def _get_halo_cached(self, halo_number) -> Halo:
        if halo_number not in self._cached_halos:
            self._cached_halos[halo_number] = self._get_halo(halo_number)
        return self._cached_halos[halo_number]

    def _get_halo(self, halo_number) -> Halo:
        halo_index = self.number_mapper.number_to_index(halo_number)
        return Halo(halo_number,
                    self._get_properties_one_halo_using_cache_if_available(halo_number, halo_index),
                    self, self.base,
                    self._get_particle_indices_one_halo_using_list_if_available(halo_number, halo_index))

    def get_dummy_halo(self, halo_number) -> DummyHalo:
        """Return a DummyHalo object containing only the halo properties, no particle information"""
        h = DummyHalo()
        h.properties.update(self.get_properties_one_halo(halo_number))
        return h

    def __len__(self) -> int:
        return self._get_num_halos()

    def __iter__(self) -> Iterable[Halo]:
        self.load_all()
        for i in self.number_mapper:
            yield self[i]

    def __repr__(self):
        return f"<{type(self).__name__}, length {len(self)}>"

    def keys(self):
        """Return an iterable of all halo numbers in the catalogue."""
        return self.number_mapper.all_numbers

    def __getitem__(self, item) -> Halo | SubhaloCatalogue:
        from .subhalo_catalogue import SubhaloCatalogue
        if isinstance(item, slice):
            return SubhaloCatalogue(self, np.arange(*item.indices(len(self))))
        elif hasattr(item, "__len__"):
            return SubhaloCatalogue(self, item)
        else:
            return self._get_halo_cached(item)

    @property
    def base(self) -> snapshot.SimSnap:
        """The snapshot object that this halo catalogue is based on."""
        return self._base()

    def _init_iord_to_fpos(self):
        """Create a member array, _iord_to_fpos, that maps particle IDs to file positions.

        This is a convenience function for subclasses to use."""
        if not hasattr(self, "_iord_to_fpos"):
            if 'iord' in self.base.loadable_keys() or 'iord' in self.base.keys():
                self._iord_to_fpos = make_iord_to_offset_mapper(self.base['iord'])

            else:
                warnings.warn("No iord array available; assuming halo catalogue is using sequential particle IDs",
                              RuntimeWarning)

                class OneToOneIndex:
                    def __getitem__(self, i):
                        return i

                self._iord_to_fpos = OneToOneIndex()

    def _get_subhalo_catalogue(self, parent_halo_number: int) -> SubhaloCatalogue:
        from .subhalo_catalogue import SubhaloCatalogue
        props = self.get_properties_one_halo(parent_halo_number)
        if 'children' in props:
            return SubhaloCatalogue(self, props['children'])
        else:
            raise ValueError(f"This halo catalogue does not support subhalos")

    @util.deprecated("This method is deprecated and will be removed in a future release. Use python `in` syntax instead.")
    def contains(self, halo_number: int) -> bool:
        """Deprecated alias; instead of ``h.contains(number)`` use ``number in h``."""
        return halo_number in self

    def __contains__(self, halo_number) -> bool:
        """Returns True if the halo catalogue contains the specified halo number."""
        return halo_number in self.number_mapper

    def get_group_array(self, family=None, use_index=False):
        """Return an array with an integer for each particle in the simulation, indicating the halo of that particle.

        If there are multiple levels (i.e. subhalos), the number returned corresponds to the lowest level, i.e.
        the smallest subhalo.

        Parameters
        ----------

        family : str, optional
            If specified, return only the group array for the specified family.

        use_index: bool, optional
            If True, return the halo index rather than the halo number. (See the class documentation for the
            distinction between halo numbers and indices.)

        """
        self.load_all()
        number_per_particle = self._index_lists.get_halo_number_per_particle(len(self.base),
                                                                             None if use_index else self.number_mapper)
        if family is not None:
            return number_per_particle[self.base._get_family_slice(family)]
        else:
            return number_per_particle

    def load_copy(self, halo_number):
        """Load a fresh SimSnap with only the particles in specified halo

        This relies on the underlying SimSnap being capable of partial loading."""
        from .. import load
        halo_index = self.number_mapper.number_to_index(halo_number)
        return load(self.base.filename,
                    take=self._get_particle_indices_one_halo_using_list_if_available(halo_number, halo_index))

    def physical_units(self, distance='kpc', velocity='km s^-1', mass='Msol', persistent=True, convert_parent=False):
        self.base.physical_units(distance=distance, velocity=velocity, mass=mass, persistent=persistent)

        # Convert all instantiated subhalos
        for halo in self._cached_halos.values():
            halo.physical_units(
                distance,
                velocity,
                mass,
                persistent=persistent,
                convert_parent=False
            )

        all_units = [units.Unit(x) for x in (distance, velocity, mass, 'a', 'h', 'K')]

        if persistent:
            self._persistent_units = all_units

        self._cached_properties_to_physical_units(all_units)

    def _cached_properties_to_physical_units(self, all_units):
        if self._properties is not None:
            for k in self._properties:
                if isinstance(self._properties[k], array.SimArray) and units.has_unit(self._properties[k]):
                    self.base._autoconvert_array_unit(self._properties[k], all_units)


    @classmethod
    def _can_load(cls, sim):
        return False

from . import (
    adaptahop,
    ahf,
    hbtplus,
    hop,
    number_array,
    rockstar,
    subfind,
    subfindhdf,
    velociraptor,
)


def _fix_american_spelling(p):
    """Map American to British spelling (used by SimSnap.halos to allow flexible spelling)"""
    if isinstance(p, str) and p.endswith('Catalog'):
        return p.replace('Catalog', 'Catalogue')
    else:
        return p
def _alias_american_spelling():
    """Create American spelling aliases for all HaloCatalogue subclasses."""
    for c in HaloCatalogue.iter_subclasses():
        american_name = c.__name__.replace("Catalogue", "Catalog")
        # put american_name into the same module as c (not this module)

        if c.__module__.startswith('pynbody.halo.'):
            module = eval(c.__module__.replace('pynbody.halo.', ''))

        setattr(module, american_name, c)

    globals()['HaloCatalog'] = HaloCatalogue

_alias_american_spelling()



"""Implementation of the key simulation snapshot class.

In a sense this is the core of pynbody, as it is the class that manages data and is subclasses to provide
IO for different simulation formats.

For an introduction, see :doc:`/tutorials/data_access`."""

from __future__ import annotations

import copy
import gc
import hashlib
import logging
import pathlib
import re
import threading
import traceback
import typing
import warnings
import weakref
from functools import reduce
from typing import Iterator

import numpy as np

from .. import (
    array,
    dependencytracker,
    family,
    filt,
    simdict,
    transformation,
    units,
    util,
)
from ..units import has_units
from ..util import iter_subclasses
from .util import ContainerWithPhysicalUnitsOption

if typing.TYPE_CHECKING:
    from .. import bridge, halo, subsnap, transformation


logger = logging.getLogger('pynbody.snapshot.simsnap')

class SimSnap(ContainerWithPhysicalUnitsOption, iter_subclasses.IterableSubclasses, transformation.Transformable):

    """The class for managing simulation snapshots.

    For most purposes, SimSnaps should be initialized through :func:`~pynbody.snapshot.load`
    or :func:`~pynbody.snapshot.new`.

    For an introduction to using this class, see :doc:`/tutorials/data_access`.

    """

    _derived_array_registry = {}

    _decorator_registry = {}

    _loadable_keys_registry = {}
    _persistent = ["kdtree", "_immediate_cache", "_kdtree_derived_smoothing"]

    # These 3D arrays get four views automatically created, one reflecting the
    # full Nx3 data, the others reflecting Nx1 slices of it
    #
    # TO DO: This should probably be read in from a config file
    _split_arrays = {'pos': ('x', 'y', 'z'),
                     'vel': ('vx', 'vy', 'vz')}

    @classmethod
    def _array_name_1D_to_ND(self, name):
        """Map a 1D array name to a corresponding 3D array name, or return None
        if no such mapping is possible.

        e.g. 'vy' -> 'vel'; 'acc_z' -> 'acc'; 'mass' -> None"""
        for k, v in self._split_arrays.items():
            if name in v:
                return k

        generic_match = re.findall("^(.+)_[xyz]$", name)
        if len(generic_match) == 1 and generic_match[0] not in self._split_arrays:
            return generic_match[0]

        return None

    @classmethod
    def _array_name_ND_to_1D(self, array_name):
        """Give the 3D array names derived from a 3D array.

        This routine makes no attempt to establish whether the array
        name passed in should indeed be a 3D array. It just returns
        the 1D slice names on the assumption that it is. This is an
        important distinction between this procedure and the reverse
        mapping as implemented by _array_name_1D_to_ND."""

        if array_name in self._split_arrays:
            array_name_1D = self._split_arrays[array_name]
        else:
            array_name_1D = [array_name + "_" + i for i in ('x', 'y', 'z')]

        return array_name_1D

    def _array_name_implies_ND_slice(self, array_name):
        """Returns True if, at best guess, the array name corresponds to a 1D slice
        of a ND array, on the basis of names alone.

        This routine first looks at special cases (pos -> x,y,z for example),
        then looks for generic names such as acc_x - however this would only be
        considered a "match" for a ND subslice if 'acc' is in loadable_keys().
        """
        for v in self._split_arrays.values():
            if array_name in v:
              return True

        generic_match = re.findall("^(.+)_[xyz]$", array_name)
        loadable_keys = self.loadable_keys()
        keys = list(self.keys())
        if len(generic_match) == 1 and generic_match[0] not in self._split_arrays:
            return generic_match[0] in loadable_keys or generic_match[0] in keys
        return False

    @classmethod
    def _can_load(cls, filepath: pathlib.Path):
        # this should be implemented by subclasses that can load from disk
        return False


    def __init__(self):
        """Initialize an empty, zero-length SimSnap.

        For most purposes SimSnaps should instead be initialized through :func:`~pynbody.snapshot.load`
        or :func:`~pynbody.snapshot.new`.
        """

        super().__init__()

        self._arrays = {}
        self._num_particles = 0
        self._family_slice = {}
        self._family_arrays = {}
        self._derived_array_names = []
        self._family_derived_array_names = {}
        self._get_array_lock = threading.RLock()
        for i in family._registry:
            self._family_derived_array_names[i] = []

        self._dependency_tracker = dependencytracker.DependencyTracker()
        self._immediate_cache_lock = threading.RLock()

        self._persistent_objects = {}

        self._unifamily = None

        # If True, when new arrays are created they are in shared memory by
        # default
        self._shared_arrays = False

        self.lazy_off = util.ExecutionControl()
        # use 'with lazy_off :' blocks to disable all hidden/lazy behaviour

        self.lazy_derive_off = util.ExecutionControl()
        # use 'with lazy_derive_off : ' blocks to disable lazy-derivation

        self.lazy_load_off = util.ExecutionControl()
        # use 'with lazy_load_off : ' blocks to disable lazy-loading

        self.auto_propagate_off = util.ExecutionControl()
        # use 'with auto_propagate_off : ' blocks to disable auto-flagging changes
        # (i.e. prevent lazy-evaluated arrays from auto-re-evaluating when their
        # dependencies change)

        self.immediate_mode = util.ExecutionControl()
        # use 'with immediate_mode: ' to always return actual numpy arrays, rather
        # than IndexedSubArrays which point to sub-parts of numpy arrays
        self.immediate_mode.on_exit = lambda: self._clear_immediate_mode()

        self.delay_promotion = util.ExecutionControl()
        # use 'with delay_promotion: ' to prevent any family arrays being promoted
        # into simulation arrays (which can cause confusion because the array returned
        # from create_family_array might have properties you don't expect)

        self.delay_promotion.on_exit = lambda: self._delayed_array_promotions(
        )
        self.__delayed_promotions = []



        self.properties = simdict.SimDict({})
        self._file_units_system = []

    ############################################
    # THE BASICS: SIMPLE INFORMATION
    ############################################

    @property
    def filename(self):
        """The filename of the snapshot, if it was loaded from disk."""
        return self._filename

    def __len__(self) -> int:
        return self._num_particles

    def __repr__(self):
        if self._filename != "":
            return "<SimSnap \"" + str(self._filename) + "\" len=" + str(len(self)) + ">"
        else:
            return "<SimSnap len=" + str(len(self)) + ">"

    def families(self) -> list[family.Family]:
        """Return the particle families which have representitives in this SimSnap.
        The families are ordered by their appearance in the snapshot."""
        out = []
        start = {}
        for fam in family._registry:
            sl = self._get_family_slice(fam)
            if sl.start != sl.stop:
                out.append(fam)
                start[fam] = (sl.start)
        out.sort(key=start.__getitem__)
        return out

    ############################################
    # THE BASICS: GETTING AND SETTING
    ############################################

    def __getitem__(self, i) -> array.SimArray | subsnap.SubSnapBase :
        """Return either a specific array or a subview of this simulation. See
        the class documentation (:class:`SimSnap`) for more information."""
        from . import subsnap

        if isinstance(i, str):
            return self._get_array_with_lazy_actions(i)
        elif isinstance(i, slice):
            return subsnap.SubSnap(self, i)
        elif isinstance(i, family.Family):
            return subsnap.FamilySubSnap(self, i)
        elif isinstance(i, np.ndarray) and np.issubdtype(np.bool_, i.dtype):
            return self._get_subsnap_from_mask_array(i)
        elif isinstance(i, (list, tuple, np.ndarray, filt.Filter)):
            return subsnap.IndexedSubSnap(self, i)
        elif isinstance(i, int) or isinstance(i, np.int32) or isinstance(i, np.int64):
            return subsnap.IndexedSubSnap(self, (i,))

        raise TypeError

    def __setitem__(self, name, item):
        """Set the contents of an array in this snapshot"""
        if self.is_derived_array(name) and not self.auto_propagate_off:
            raise RuntimeError("Derived array is not writable")

        if isinstance(name, tuple) or isinstance(name, list):
            index = name[1]
            name = name[0]
        else:
            index = None

        self._assert_not_family_array(name)

        if isinstance(item, array.SimArray) or isinstance(item, array.IndexedSimArray):
            ax = item
        else:
            ax = np.asanyarray(item).view(array.SimArray)

        if name not in list(self.keys()):
            # Array needs to be created. We do this through the
            # private _create_array method, so that if we are operating
            # within a particle-specific subview we automatically create
            # a particle-specific array
            try:
                ndim = len(ax[0])
            except TypeError:
                ndim = 1
            except IndexError:
                ndim = ax.shape[-1] if len(ax.shape) > 1 else 1

            # The dtype will be the same as an existing family array if
            # one exists, or the dtype of the source array we are copying
            dtype = self._get_preferred_dtype(name)
            if dtype is None:
                dtype = getattr(item, 'dtype', None)

            self._create_array(name, ndim, dtype=dtype)

        # Copy in contents if the contents isn't actually pointing to
        # the same data (which will be the case following operations like
        # += etc, since these call __setitem__).
        self._set_array(name, ax, index)

    def __delitem__(self, name):
        if name in self._family_arrays:
            # mustn't have simulation-level array of this name
            assert name not in self._arrays
            del self._family_arrays[name]

            for v in self._family_derived_array_names.values():
                if name in v:
                    del v[v.index(name)]

        else:
            del self._arrays[name]
            if name in self._derived_array_names:
                del self._derived_array_names[
                    self._derived_array_names.index(name)]


    def _get_subsnap_from_mask_array(self,mask_array):
        if len(mask_array.shape) > 1 or mask_array.shape[0] > len(self):
            raise ValueError("Incorrect shape for masking array")
        else:
            return self[np.where(mask_array)]

    def _get_array_with_lazy_actions(self, name):
        """Get an array by the given name; if it's not available, attempt to load or derive it first"""

        # the below is not thread-safe, so we lock
        with self._get_array_lock:
            if name in list(self.keys()):
                self._dependency_tracker.touching(name)

                # Ensure that any underlying dependencies on 1D positions and velocities
                # are forwarded to 3D dependencies as well
                nd_name = self._array_name_1D_to_ND(name)
                if nd_name is not None:
                    self._dependency_tracker.touching(nd_name)

            elif not self.lazy_off:
                # The array is not currently in memory at the level we need it, and there is a possibility
                # of getting it into memory using lazy derivation or loading. First, if there is a family level
                # array by the same name, dispose of it. (Note if this is being called on a FamilySubSnap, the below
                # has no effect, and anyway we wouldn't reach this point in the code if the family array were available.)
                self.__resolve_obscuring_family_array(name)

                # Now, we'll try to load the array...
                if not self.lazy_load_off:
                    # Note that we don't want this to be inside _dependency_tracker.calculating(name), because there is a
                    # small possibility the load will be mapped into a derivation by the loader class. Specifically this
                    # happens in ramses snapshots for the mass array (which is derived from the density array for gas cells).
                    self.__load_if_required(name)

                if name in self:
                    # We managed to load it. Note the dependency.
                    self._dependency_tracker.touching(name)
                elif not self.lazy_derive_off:
                    # Try deriving instead
                    with self._dependency_tracker.calculating(name):
                        self.__derive_if_required(name)

        # At this point we've done everything we can to get the array into memory. If it's still not there, we'll
        # get a KeyError from the below.
        return self._get_array(name)



    def __load_if_required(self, name):
        if name not in list(self.keys()):
            try:
                self.__load_array_and_perform_postprocessing(name)
            except OSError:
                pass

    def __derive_if_required(self, name):
        if name not in list(self.keys()):
            self._derive_array(name)

    def __resolve_obscuring_family_array(self, name):
        """When a lazy derive or lazy load is about to happen for array given by name, deal with family arrays by the same name

        Note that if this is being called from a family subview, family_keys() returns an empty set by definition, so
        nothing bad happens. It is when a derivation/lazy-load is attempted across the snapshot that having family arrays
        of the same name might get in the way.
        """
        if name in self.family_keys():
            self.__remove_family_array_if_derived(name)

        if name in self.family_keys():
            self.__load_remaining_families_if_loadable(name)

        if name in self.family_keys():
            in_fam, out_fam = self.__get_included_and_excluded_families_for_array(name)
            raise KeyError("""{!r} is a family-level array for {}. To use it over the whole simulation you need either to delete it first, or create it separately for {}.""".format(
                name, in_fam, out_fam))

    def __get_included_and_excluded_families_for_array(self,name):
        in_fam = []
        out_fam = []
        for x in self.families():
            if name in self[x]:
                in_fam.append(x)
            else:
                out_fam.append(x)

        return in_fam, out_fam

    def __remove_family_array_if_derived(self, name):
        if self.is_derived_array(name):
            del self.ancestor[name]


    def __load_remaining_families_if_loadable(self, name):
        in_fam, out_fam = self.__get_included_and_excluded_families_for_array(name)
        try:
            for fam in out_fam:
                if name not in self[fam]: # (check again, in case it was created in the meantime)
                    self.__load_array_and_perform_postprocessing(name, fam=fam)
        except OSError:
            pass



    def _get_persist(self, hash, name):
        try:
            return self._persistent_objects[hash][name]
        except Exception:
            return None

    def _set_persist(self, hash, name, obj=None):
        if hash not in self._persistent_objects:
            self._persistent_objects[hash] = {}
        self._persistent_objects[hash][name] = obj

    def _clear_immediate_mode(self):
        for k, v in self._persistent_objects.items():
            if '_immediate_cache' in v:
                del v['_immediate_cache']

    def __getattr__(self, name):
        """This function overrides the behaviour of f.X where f is a SimSnap object.

        It serves two purposes; first, it provides the family-handling behaviour
        which makes f.dm equivalent to f[pynbody.family.dm]. Second, it implements
        persistent objects -- properties which are shared between two equivalent SubSnaps."""
        if name in SimSnap._persistent:
            obj = self.ancestor._get_persist(self._inclusion_hash, name)
            if obj:
                return obj

        try:
            return self[family.get_family(name)]
        except ValueError:
            pass

        raise AttributeError("{!r} object has no attribute {!r}".format(
            type(self).__name__, name))

    def __setattr__(self, name, val):
        """This function overrides the behaviour of setting f.X where f is a SimSnap object.

        It serves two purposes; first it prevents overwriting of family names (so you can't
        write to, for instance, f.dm). Second, it implements persistent objects -- properties
        which are shared between two equivalent SubSnaps."""
        if name in family.family_names():
            raise AttributeError("Cannot assign family name " + name)

        if name in SimSnap._persistent:
            self.ancestor._set_persist(self._inclusion_hash, name, val)
        else:
            return object.__setattr__(self, name, val)

    def __delattr__(self, name):
        """This function allows persistent objects (as shared between two equivalent SubSnaps)
        to be permanently deleted."""
        if name in SimSnap._persistent:
            obj = self.ancestor._get_persist(self._inclusion_hash, name)
            if obj:
                self.ancestor._set_persist(self._inclusion_hash, name, None)
                try:
                    object.__delattr__(self, name)
                except AttributeError:
                    pass
                return
        object.__delattr__(self, name)

    ############################################
    # DICTIONARY EMULATION FUNCTIONS
    ############################################
    def keys(self) -> list[str]:
        """Return the directly accessible array names (in memory)"""
        return list(self._arrays.keys())

    def has_key(self, name: str) -> bool:
        """Returns True if the array name is accessible (in memory)"""
        return name in self.keys()

    def values(self) -> list[array.SimArray]:
        """Returns a list of the actual arrays in memory"""
        x = []
        for k in list(self.keys()):
            x.append(self[k])
        return x

    def items(self) -> list[tuple[str, array.SimArray]]:
        """Returns a list of tuples describing the array
        names and their contents in memory"""
        x = []
        for k in list(self.keys()):
            x.append((k, self[k]))
        return x

    def get(self, key, alternative=None) -> array.SimArray | None :
        """Standard python get method, returns self[key] if
        key in self else alternative"""
        try:
            return self[key]
        except KeyError:
            return alternative

    def iterkeys(self) -> Iterator[str]:
        """Iterator over the array names in memory"""
        yield from list(self.keys())

    __iter__ = iterkeys

    def itervalues(self) -> Iterator[array.SimArray]:
        """Iterator over the arrays in memory"""
        for k in self:
            yield self[k]

    def iteritems(self) -> Iterator[tuple[str, array.SimArray]]:
        """Iterator over the array names and their contents in memory"""
        for k in self:
            yield (k, self[k])

    ############################################
    # DICTIONARY-LIKE FUNCTIONS
    # (not in the normal interface for dictionaries,
    # but serving similar purposes)
    ############################################

    def has_family_key(self, name: str) -> bool:
        """Returns True if the array name is accessible (in memory) for at least one family"""
        return name in self.family_keys()

    def loadable_keys(self, fam=None) -> list[str]:
        """Returns a list of arrays which can be lazy-loaded from
        an auxiliary file."""
        return []

    def derivable_keys(self) -> list[str]:
        """Returns a list of arrays which can be lazy-evaluated."""
        res = []
        for cl in type(self).__mro__:
            if cl in self._derived_array_registry:
                res += list(self._derived_array_registry[cl].keys())
        return res

    def all_keys(self) -> list[str]:
        """Returns a list of all arrays that can be either lazy-evaluated
        or lazy loaded from an auxiliary file."""
        return self.derivable_keys() + self.loadable_keys()

    def family_keys(self, fam=None) -> list[str]:
        """Return list of arrays which are not accessible from this
        view, but can be accessed from family-specific sub-views.

        If *fam* is not None, only those keys applying to the specific
        family will be returned (equivalent to self.fam.keys())."""
        if fam is not None:
            return [x for x in self._family_arrays if fam in self._family_arrays[x]]
        else:
            return list(self._family_arrays.keys())

    ############################################
    # ANCESTRY FUNCTIONS
    ############################################

    def is_ancestor(self, other) -> bool:
        """Returns true if other is a subview of self"""

        if other is self:
            return True
        elif hasattr(other, 'base'):
            return self.is_ancestor(other.base)
        else:
            return False

    def is_descendant(self, other) -> bool:
        """Returns true if self is a subview of other"""
        return other.is_ancestor(self)

    @property
    def ancestor(self) -> SimSnap:
        """The original SimSnap from which this view is derived (potentially self)"""
        if hasattr(self, 'base'):
            return self.base.ancestor
        else:
            return self

    def get_index_list(self, relative_to, of_particles=None) -> np.ndarray:
        """Get a list specifying the index of the particles in this view relative to the ancestor *relative_to*

        The returned index list satisfies ``relative_to[get_index_list(relative_to)]==self``."""

        # Implementation for base snapshot

        if self is not relative_to:
            raise RuntimeError("Not a descendant of the specified simulation")
        if of_particles is None:
            of_particles = np.arange(len(self))

        return of_particles

    ############################################
    # SET-LIKE OPERATIONS FOR SUBSNAPS
    ############################################
    def intersect(self, other, op=np.intersect1d) -> SimSnap:
        """Returns the set intersection of this simulation view with another view
        of the same simulation"""

        anc = self.ancestor
        if not anc.is_ancestor(other):
            raise RuntimeError("Parentage is not suitable")

        a = self.get_index_list(anc)
        b = other.get_index_list(anc)
        return anc[op(a, b)]

    def union(self, other) -> SimSnap:
        """Returns the set union of this simulation view with another view
        of the same simulation"""

        return self.intersect(other, op=np.union1d)

    def setdiff(self, other) -> SimSnap:
        """Returns the set difference of this simulation view with another view
        of the same simulation"""

        return self.intersect(other, op=np.setdiff1d)

    ############################################
    # UNIT MANIPULATION
    ############################################
    def conversion_context(self) -> dict[str, float]:
        """Return a dictionary containing a (scalefactor) and h
        (Hubble constant in canonical units) for this snapshot, ready for
        passing into unit conversion functions."""
        d = {}
        wanted = ['a', 'h']
        for x in wanted:
            if x in self.properties:
                d[x] = self.properties[x]
        return d

    def _override_units_system(self):
        """Look for and process a text file with a custom units system for this snapshot.

        The text file should be named <filename>.units and contain unit specifications, one-per-line, e.g.

        pos: kpc a
        vel: km s^-1
        mass: Msol

        This override functionality needs to be explicitly called by a subclass after it has initialised
        its best guess at the units.
        """
        try:
            with open(str(self.filename) + ".units") as f:
                lines = f.readlines()
        except OSError:
            return

        name_mapping = {'pos': 'distance', 'vel': 'velocity'}
        units_dict = {}

        for line in lines:
            if (not line.startswith("#")):
                if ":" not in line:
                    raise OSError("Unknown format for units file %r"%(self.filename+".units"))
                else:
                    t, u = list(map(str.strip,line.split(":")))
                    t = name_mapping.get(t,t)
                    units_dict[t] = u

        self.set_units_system(**units_dict)

    def set_units_system(self, velocity=None, distance=None, mass=None, temperature=None):
        """Set the unit system for the snapshot by specifying any or
        all of `velocity`, `distance`, `mass` and `temperature`
        units. The units can be given as strings or as pynbody `Unit`
        objects.

        If any of the units are not specified and a previous
        `file_units_system` does not exist, the defaults are used.
        """
        import configparser

        from .. import config_parser

        # if the units system doesn't exist (if this is a new snapshot), create
        # one
        if len(self._file_units_system) < 3:
            warnings.warn("Previous unit system incomplete -- using defaults")
            self._file_units_system = [
                units.Unit(x) for x in ('G', '1 kpc', '1e10 Msol')]

        else:
            # we want to change the base units -- so convert to original
            # units first and then set all arrays to new unit system
            self.original_units()


        # if any are missing, work them out from what we already have:

        if velocity is None:
            velocity = self.infer_original_units('km s^-1')

        if distance is None:
            distance = self.infer_original_units('kpc')

        if mass is None:
            mass = self.infer_original_units('Msol')

        if temperature is None:
            temperature = self.infer_original_units('K')

        new_units = []
        for x in [velocity, distance, mass, temperature]:
            if x is not None:
                new_units.append(units.Unit(x))


        self._file_units_system = new_units

        # set new units for all known arrays
        for arr_name in list(self.keys()):
            arr = self[arr_name]
            # if the array has units, then use the current units, else
            # check if a default dimension for this array exists in
            # the configuration
            if arr.units != units.NoUnit():
                ref_unit = arr.units
            else:
                try:
                    ref_unit = config_parser.get(
                        'default-array-dimensions', arr_name)
                except configparser.NoOptionError:
                    # give up -- no applicable dimension found
                    continue

            arr.set_units_like(ref_unit)

    def original_units(self):
        """Converts all arrays'units to be consistent with the units of
        the original file."""
        self.physical_units(distance=self.infer_original_units('km'),
                            velocity=self.infer_original_units('km s^-1'),
                            mass=self.infer_original_units('Msol'), persistent=False)


    def infer_original_units(self, dimensions):
        """Given a unit (or string) `dimensions`, returns a unit with the same
        physical dimensions which is in the unit schema of the current file."""
        dimensions = units.Unit(dimensions)
        try:
            d = dimensions.dimensional_project(
                self._file_units_system + ["a", "h"])
        except units.UnitsException:
            return units.no_unit
        new_unit = reduce(lambda x, y: x * y, [
                          a ** b for a, b in zip(self._file_units_system, d)])
        return new_unit

    def _default_units_for(self, array_name):
        """Attempt to construct and return the units for the named array
        on disk, using what we know about the purpose of arrays (in config.ini)
        and the original unit system (via infer_original_units)."""
        array_name = self._array_name_1D_to_ND(array_name) or array_name
        u = units._default_units.get(array_name, None)
        if u is not None:
            u = self.infer_original_units(u)
        return u

    def halos(self, *args, **kwargs) -> halo.HaloCatalogue:
        """Tries to instantiate a halo catalogue object for the given snapshot.

        For introductory information about halo catalogues, see :ref:`halo_tutorial`, and the documentation for
        :py:mod:`pynbody.halo`.

        Multiple catalogue classes are available in pynbody, and they are tried in turn until the first which
        accepted the request to load halos for this file.

        The order of catalogue class priority is either defined in the configuration file or can be passed
        as a 'priority' keyword argument, which should be a list of either class names or classes.

        For example

        >>> h = snap.halos(priority = ["HOPCatalogue", "AHFCatalogue", pynbody.halo.subfind.SubfindCatalogue])

        would try to load HOP halos, then AHF halos, then Subfind halos, before finally trying all other
        known halo classes in an arbitrary order.

        Arguments and keyword arguments other than `priority` are passed onto the individual halo loaders.
        If a given catalogue class does not accept the args/kwargs that you pass in, by definition it cannot
        be used; this can lead to 'silent' failures. To understand why a given class is not being instantiated
        by this method, the best option is to try _directly_ instantiating that class to reveal the error
        explicitly.
        """

        from .. import config, halo

        priority = kwargs.pop('priority',
                              config['halo-class-priority'])

        priority = [halo._fix_american_spelling(p) for p in priority]

        for c in halo.HaloCatalogue.iter_subclasses_with_priority(priority):
            try:
                can_load = c._can_load(self, *args, **kwargs)
            except TypeError:
                can_load = False

            if can_load:
                return c(self, *args, **kwargs)

        raise RuntimeError("No halo catalogue found for %r" % str(self))

    def bridge(self, other) -> bridge.AbstractBridge:
        """Tries to construct a bridge function between this SimSnap
        and another one.

        This function calls :func:`pynbody.bridge.bridge_factory`. For
        more information see :ref:`bridge-tutorial`, or the reference
        documentation for :py:mod:`pynbody.bridge`.

        """
        from .. import bridge
        return bridge.bridge_factory(self, other)

    def load_copy(self) -> SimSnap:
        """Tries to load a copy of this snapshot, using partial loading to select
        only a subset of particles corresponding to a given SubSnap"""
        from .. import load
        if getattr(self.ancestor,'partial_load',False):
            raise NotImplementedError("Cannot load a copy of data that was itself partial-loaded")
        return load(self.ancestor.filename, take=self.get_index_list(self.ancestor))

    ############################################
    # HELPER FUNCTIONS FOR LAZY LOADING
    ############################################
    def _load_array(self, array_name, fam=None):
        """This function is called by the framework to load an array
        from disk and should be overloaded by child classes.

        If *fam* is not None, the array should be loaded only for the
        specified family.
        """
        raise OSError("No lazy-loading implemented")

    def __load_array_and_perform_postprocessing(self, array_name, fam=None):
        """Calls _load_array for the appropriate subclass, but also attempts to convert
        units of anything that gets loaded and automatically loads the whole ND array
        if this is a subview of an ND array"""
        array_name = self._array_name_1D_to_ND(array_name) or array_name

        # keep a record of every array in existence before load (in case it
        # triggers loading more than we expected, e.g. coupled pos/vel fields
        # etc)
        anc = self.ancestor

        pre_keys = set(anc.keys())

        # the following function builds a dictionary mapping families to a set of the
        # named arrays defined for them.
        fk = lambda: {fami: {k for k in list(anc._family_arrays.keys()) if fami in anc._family_arrays[k]}
                           for fami in family._registry}
        pre_fam_keys = fk()

        with self.delay_promotion:
            # delayed promotion is required here, otherwise units get messed up when
            # a simulation array gets promoted mid-way through our loading process.
            #
            # see the gadget unit test, test_unit_persistence
            if fam is not None:
                self._load_array(array_name, fam)
            else:
                try:
                    self._load_array(array_name, fam)
                except OSError:
                    for fam_x in self.families():
                        self._load_array(array_name, fam_x)

            # Find out what was loaded
            new_keys = set(anc.keys()) - pre_keys
            new_fam_keys = fk()
            for fami in new_fam_keys:
                new_fam_keys[fami] = new_fam_keys[fami] - pre_fam_keys[fami]

            with self.lazy_off:
                # If the loader hasn't given units already, try to determine the defaults
                # Then, attempt to convert what was loaded into friendly units
                for v in new_keys:
                    if not units.has_units(anc[v]):
                        anc[v].units = anc._default_units_for(v)
                    anc._autoconvert_array_unit(anc[v])
                for f, vals in new_fam_keys.items():
                    for v in vals:
                        if not units.has_units(anc[f][v]):
                            anc[f][v].units = anc._default_units_for(v)
                        anc._autoconvert_array_unit(anc[f][v])



    ############################################
    # VECTOR TRANSFORMATIONS OF THE SNAPSHOT
    ############################################

    def wrap(self, boxsize=None, convention='center'):
        """Wraps the positions of the particles in the box to lie in a chosen fundamental domain.

        Parameters
        ----------
        boxsize : float | units.UnitBase, optional
            The size of the box to wrap the particles in. If not specified, the boxsize is taken from the snapshot's
            properties.

        convention : str, optional
            The convention to use for wrapping the particles. The default is 'center', which wraps the particles to lie
            in the range [-boxsize/2, boxsize/2). Alternatively, 'upper' wraps the particles to lie in the range
            [0, boxsize).
        """


        if boxsize is None:
            boxsize = self.properties["boxsize"]

        if isinstance(boxsize, units.UnitBase):
            boxsize = float(boxsize.ratio(self["pos"].units, **self.conversion_context()))

        if convention=='center':
            for coord in "x", "y", "z":
                self[coord][np.where(self[coord] <= -boxsize / 2)] += boxsize
                self[coord][np.where(self[coord] > boxsize / 2)] -= boxsize
        elif convention=='upper':
            for coord in "x", "y", "z":
                self[coord][np.where(self[coord] < 0)] += boxsize
                self[coord][np.where(self[coord] >= boxsize)] -= boxsize
        else:
            raise ValueError("Unknown wrapping convention")

    ############################################
    # WRITING FUNCTIONS
    ############################################
    def write(self, fmt=None, filename=None, **kwargs):
        if filename is None and "<" in self.filename:
            raise RuntimeError(
                'Cannot infer a filename; please provide one (use obj.write(filename="filename"))')

        if fmt is None:
            if not hasattr(self, "_write"):
                raise RuntimeError(
                    'Cannot infer a file format; please provide one (e.g. use obj.write(filename="filename", fmt=pynbody.snapshot.tipsy.TipsySnap)')

            self._write(self, filename, **kwargs)
        else:
            fmt._write(self, filename, **kwargs)

    def write_array(self, array_name, fam=None, overwrite=False, **kwargs):
        """
        Write out the array with the specified name.

        Only a subset of SimSnap subclasses support this operation. If the subclass does not support it, an ``OSError``
        will be raised.

        .. seealso:: :meth:`pynbody.array.SimArray.write`

        Parameters
        ----------

        array_name : str
            the name of the array to write

        fam : str or list of str, optional
            Write out only one family; or provide a list to write out a set of families.

        overwrite : bool, optional
            If True, overwrite the array on disk if it already exists. If False (default), an OSError will be raised
            if the array already exists.


        """

        # Determine whether this is a write or an update
        if fam is None:
            fam = self.families()

        # It's an update if we're not fully replacing the file on
        # disk, i.e. there exists a family f in self.families() but
        # not in fam for which array_name is loadable
        is_update = any([array_name in self[
                        f].loadable_keys() and f not in fam for f in self.families()])

        if not hasattr(self, "_write_array"):
            raise OSError(
                "The underlying file format class does not support writing individual arrays back to disk")

        if is_update and not hasattr(self, "_update_array"):
            raise OSError(
                "The underlying file format class does not support updating arrays on disk")

        # It's an overwrite if we're writing over something loadable
        is_overwriting = any([array_name in self[
                             f].loadable_keys() for f in fam])

        if is_overwriting and not overwrite:
            # User didn't specifically say overwriting is OK
            raise OSError(
                "This operation would overwrite existing data on disk. Call again setting overwrite=True if you want to enable this behaviour.")

        if is_update:
            self._update_array(array_name, fam=fam, **kwargs)
        else:
            self._write_array(self, array_name, fam=fam, **kwargs)

    ############################################
    # LOW-LEVEL ARRAY MANIPULATION
    ############################################
    def _get_preferred_dtype(self, array_name):
        """Return the 'preferred' numpy datatype for a named array.

        This is mainly useful when creating family arrays for new families, to be
        sure the datatype chosen matches"""

        if hasattr(self, 'base'):
            return self.base._get_preferred_dtype(array_name)
        elif array_name in list(self.keys()):
            return self[array_name].dtype
        elif array_name in self.family_keys():
            return self._family_arrays[array_name][list(self._family_arrays[array_name].keys())[0]].dtype
        else:
            return None

    def _create_array(self, array_name, ndim=1, dtype=None, zeros=True, derived=False, shared=None,
                      source_array=None):
        """Create a single snapshot-level array of dimension len(self) x ndim, with
        a given numpy dtype.

        *kwargs*:

          - *ndim*: the number of dimensions for each particle
          - *dtype*: a numpy datatype for the new array
          - *zeros*: if True, zeros the array (which takes a bit of time); otherwise
            the array is uninitialized
          - *derived*: if True, this new array will be flagged as a derived array
            which makes it read-only
          - *shared*: if True, the array will be built on top of a shared-memory array
            to make it possible to access from another process. If 'None' (default), shared
            memory will be used if the snapshot prefers it, otherwise not.
          - *source_array*: if provided, the SimSnap will take ownership of this specified
            array rather than create a new one
        """

        # Does this actually correspond to a slice into a 3D array?
        NDname = self._array_name_1D_to_ND(array_name)
        if NDname:
            self._create_array(
                NDname, ndim=3, dtype=dtype, zeros=zeros, derived=derived)
            return

        if ndim == 1:
            dims = (self._num_particles, )
        else:
            dims = (self._num_particles, ndim)

        if shared is None:
            shared = self._shared_arrays

        if source_array is None:
            source_array = array.array_factory(dims, dtype, zeros, shared)
        else:
            assert isinstance(source_array, array.SimArray)
            assert source_array.shape == dims

        source_array._sim = weakref.ref(self)
        source_array._name = array_name
        source_array.family = None

        self._arrays[array_name] = source_array

        if derived:
            if array_name not in self._derived_array_names:
                self._derived_array_names.append(array_name)

        if ndim == 3:
            array_name_1D = self._array_name_ND_to_1D(array_name)

            for i, a in enumerate(array_name_1D):
                self._arrays[a] = source_array[:, i]
                self._arrays[a]._name = a

    def _create_family_array(self, array_name, family, ndim=1, dtype=None, derived=False, shared=None,
                             source_array=None):
        """Create a single array of dimension len(self.<family.name>) x ndim,
        with a given numpy dtype, belonging to the specified family. For arguments
        other than *family*, see the documentation for :func:`~pynbody.snapshot.SimSnap._create_array`.

        Warning: Do not assume that the family array will be available after
        calling this funciton, because it might be a 'completion' of existing
        family arrays, at which point the routine will actually be creating
        a simulation-level array, e.g.

        sim._create_family_array('bla', dm)
        sim._create_family_array('bla', star)
        'bla' in sim.family_keys() # -> True
        'bla' in sim.keys() # -> False
        sim._create_family_array('bla', gas)
        'bla' in sim.keys() # -> True
        'bla' in sim.family_keys() # -> False

        sim[gas]['bla'] *is* guaranteed to exist, however, it just might
        be a view on a simulation-length array.

        """

        NDname = self._array_name_1D_to_ND(array_name)
        if NDname:
            self._create_family_array(
                NDname, family, ndim=3, dtype=dtype, derived=derived)
            return

        self_families = self.families()

        if len(self_families) == 1 and family in self_families:
            # If the file has only one family, just go ahead and create
            # a normal array
            self._create_array(
                array_name, ndim=ndim, dtype=dtype, derived=derived, source_array=source_array)
            return

        if ndim == 1:
            dims = (self[family]._num_particles, )
        else:
            dims = (self[family]._num_particles, ndim)

        # Determine what families already have an array of this name
        fams = []
        dtx = None
        try:
            fams = list(self._family_arrays[array_name].keys())
            dtx = self._family_arrays[array_name][fams[0]].dtype
        except KeyError:
            pass

        fams.append(family)

        if dtype is not None and dtx is not None and dtype != dtx:

            # We insist on the data types being the same for, e.g. sim.gas['my_prop'] and sim.star['my_prop']
            # This makes promotion to simulation-level arrays possible.
            raise ValueError("Requested data type {!r} is not consistent with existing data type {!r} for family array {!r}".format(
                str(dtype), str(dtx), array_name))

        if all([x in fams for x in self_families]):
            # If, once we created this array, *all* families would have
            # this array, just create a simulation-level array
            if self._promote_family_array(array_name, ndim=ndim, derived=derived, shared=shared) is not None:
                return None

        # if we get here, either the array cannot be promoted to simulation level, or that would
        # not be appropriate, so actually go ahead and create the family array

        if shared is None:
            shared = self._shared_arrays

        if source_array is None:
            source_array = array.array_factory(dims, dtype, False, shared)
        else:
            assert isinstance(source_array, array.SimArray)
            assert source_array.shape == dims
        source_array._sim = weakref.ref(self)
        source_array._name = array_name
        source_array.family = family

        def sfa(n, v):
            try:
                self._family_arrays[n][family] = v
            except KeyError:
                self._family_arrays[n] = dict({family: v})

        sfa(array_name, source_array)
        if derived:
            if array_name not in self._family_derived_array_names[family]:
                self._family_derived_array_names[family].append(array_name)

        if ndim == 3:
            array_name_1D = self._array_name_ND_to_1D(array_name)
            for i, a in enumerate(array_name_1D):
                sfa(a, source_array[:, i])
                self._family_arrays[a][family]._name = a

    def _del_family_array(self, array_name, family):
        """Delete the array with the specified name for the specified family"""
        del self._family_arrays[array_name][family]
        if len(self._family_arrays[array_name]) == 0:
            del self._family_arrays[array_name]

        derive_track = self._family_derived_array_names[family]
        if array_name in derive_track:
            del derive_track[derive_track.index(array_name)]

    def _get_from_immediate_cache(self, name, fn):
        """Retrieves the named numpy array from the immediate cache associated
        with this snapshot. If the array does not exist in the immediate
        cache, function fn is called with no arguments and must generate
        it."""

        with self._immediate_cache_lock:
            if not hasattr(self, '_immediate_cache'):
                self._immediate_cache = [{}]
            cache = self._immediate_cache[0]
            hx = hash(name)
            if hx not in cache:
                cache[hx] = fn()

        return cache[hx]

    def _get_array(self, name, index=None, always_writable=False):
        """Get the array of the specified *name*, optionally for only the particles specified by *index*.

        If *always_writable* is True, the returned array is writable. Otherwise, it is still normally writable, but
        not if the array is flagged as derived by the framework.

        Note that this routine never calls any lazy loading/derivation routines, for which one needs to call
        _get_array_with_lazy_actions.
        """

        x = self._arrays[name]
        if x.derived and not always_writable:
            x = x.view()
            x.flags['WRITEABLE'] = False

        if index is not None:
            if type(index) is slice:
                ret = x[index]
            else:
                ret = array.IndexedSimArray(x, index)

            ret.family = None
            return ret

        else:
            return x

    def _get_family_array(self, name, fam, index=None, always_writable=False):
        """Get the family-level array with specified *name* for the family *fam*,
        optionally for only the particles specified by *index* (relative to the
        family slice).

        If *always_writable* is True, the returned array is writable. Otherwise
        it is still normally writable, but not if the array is flagged as derived
        by the framework.

        Note that this routine never calls any lazy loading/derivation routines, for which one needs to call
        _get_array_with_lazy_actions on the FamilySubSnap returned by self[fam].
        """

        try:
            x = self._family_arrays[name][fam]
        except KeyError:
            raise KeyError("No array " + name + " for family " + fam.name) from None

        if x.derived and not always_writable:
            x = x.view()
            x.flags['WRITEABLE'] = False

        if index is not None:
            if type(index) is slice:
                x = x[index]
            else:
                if self.immediate_mode:
                    x = self._get_from_immediate_cache(name,
                                                       lambda: x[index])
                else:
                    x = array.IndexedSimArray(x, index)
        return x

    def _set_array(self, name, value, index=None):
        """Update the contents of the snapshot-level array to that
        specified by *value*. If *index* is not None, update only that
        subarray specified."""
        util.set_array_if_not_same(self._arrays[name], value, index)

    def _set_family_array(self, name, family, value, index=None):
        """Update the contents of the family-level array to that
        specified by *value*. If *index* is not None, update only that
        subarray specified."""
        util.set_array_if_not_same(self._family_arrays[name][family],
                                   value, index)

    def _create_arrays(self, array_list, ndim=1, dtype=None, zeros=True):
        """Create a set of arrays *array_list* of dimension len(self) x ndim, with
        a given numpy dtype."""
        for array in array_list:
            self._create_array(array, ndim, dtype, zeros)

    def _get_family_slice(self, fam):
        """Turn a specified Family object into a concrete slice which describes
        which particles in this SimSnap belong to that family."""
        try:
            return self._family_slice[fam]
        except KeyError:
            return slice(0, 0)

    def _family_index(self):
        """Return an array giving the family number of each particle in this snapshot,
        something like 0,0,0,0,1,1,2,2,2, ... where 0 means self.families()[0] etc"""

        if hasattr(self, "_family_index_cached"):
            return self._family_index_cached

        ind = np.empty((len(self),), dtype='int8')
        for i, f in enumerate(self.ancestor.families()):
            ind[self._get_family_slice(f)] = i

        self._family_index_cached = ind

        return ind

    def _assert_not_family_array(self, name):
        """Raises a ValueError if the specified array name is connected to
        a family-specific array"""
        if name in self.family_keys():
            raise KeyError("Array " + name + " is a family-level property")

    def _delayed_array_promotions(self):
        """Called automatically to catch up with pending array promotions"""
        for x in self.__delayed_promotions:
            self._promote_family_array(*x)

        self.__delayed_promotions = []

    def _promote_family_array(self, name, ndim=1, dtype=None, derived=False, shared=None):
        """Create a simulation-level array (if it does not exist) with
        the specified name. Copy in any data from family-level arrays
        of the same name."""

        if ndim == 1 and self._array_name_1D_to_ND(name):
            return self._promote_family_array(self._array_name_1D_to_ND(name), 3, dtype)

        if self.delay_promotion:
            # if array isn't already scheduled for promotion, do so now
            if not any([x[0] == name for x in self.__delayed_promotions]):
                self.__delayed_promotions.append(
                    [name, ndim, dtype, derived, shared])
            return None

        if dtype is None:
            try:
                x = list(self._family_arrays[name].keys())[0]
                dtype = self._family_arrays[name][x].dtype
                for x in list(self._family_arrays[name].values()):
                    if x.dtype != dtype:
                        warnings.warn("Data types of family arrays do not match; assuming " + str(
                            dtype),  RuntimeWarning)

            except IndexError:
                pass

        dmap = [name in self._family_derived_array_names[
            i] for i in self._family_arrays[name]]
        some_derived = any(dmap)
        all_derived = all(dmap)

        if derived:
            some_derived = True
        if not derived:
            all_derived = False

        if name not in self._arrays:
            self._create_array(
                name, ndim=ndim, dtype=dtype, derived=all_derived, shared=shared)
        try:
            for fam in self._family_arrays[name]:
                if has_units(self._family_arrays[name][fam]) and not has_units(self._arrays[name]):
                    self._arrays[name].units = self._family_arrays[
                        name][fam].units
                    # inherits the units from the first dimensional family array found.
                    # Note that future copies, once the units are set, invoke the correct conversion
                    # and raise a UnitsException if such a conversion is
                    # impossible.

                try:
                    self._arrays[name][self._get_family_slice(
                        fam)] = self._family_arrays[name][fam]
                except units.UnitsException:
                    # There is a problem getting everything into the same units. The trouble is
                    # that having got here if we let the exception propagate, we're going to
                    # end up with the SimSnap in an inconsistent state. So force the copy
                    # ignoring the units and raise a warning
                    warnings.warn(
                        "When conjoining family arrays to create a snapshot level array, the units could not be unified. You will now have a snapshot-level array %r with inconsistent unit information" % name)
                    self._arrays[name].base[self._get_family_slice(
                        fam)] = self._family_arrays[name][fam].base

            del self._family_arrays[name]
            if ndim == 3:
                for v in self._array_name_ND_to_1D(name):
                    del self._family_arrays[v]
            gc.collect()

        except KeyError:
            pass

        if some_derived:
            if all_derived:
                self._derived_array_names.append(name)
            else:
                warnings.warn(
                    "Conjoining derived and non-derived arrays. Assuming result is non-derived, so no further updates will be made.", RuntimeWarning)
            for v in self._family_derived_array_names.values():
                if name in v:
                    del v[v.index(name)]

        return self._arrays[name]

    ############################################
    # DERIVED ARRAY SYSTEM
    ############################################
    @classmethod
    def derived_array(cls, fn):
        """Function decorator to register a derivable quantity for all SimSnaps, or for a specific subclass.

        Example usage:

        >>> @pynbody.derived_array
        ... def radius_squared(sim):
        ...     return sim['x']**2 + sim['y']**2 + sim['z']**2

        When a user tries to access the ``radius_squared`` array from any snapshot, the function will be called
        and the result will be stored then returned. The function should take a single argument, the simulation object,
        and return a numpy array or SimArray of the same length as the simulation.

        The array will be automatically updated whenever arrays it relies upon for derivation are changed.

        If instead of ``pynbody.derived_array``, a derived array associated with a specific subclass of
        :class:`~pynbody.snapshot.SimSnap` is desired, one may use the following syntax:

        >>> @MySimSnapSubclass.derived_array
        ... def radius_squared(sim):
        ...     return sim['x']**2 + sim['y']**2 + sim['z']**2

        If the function has a docstring, it will be prepended with "Derived: " to indicate that it is a derived
        quantity. If the derived array is being added for a specific subclass, a list of subclasses is also
        included in the automatic modifications to the docstring.

        For more information about derived quantities see :ref:`derived-quantities`.

        .. versionchanged:: 2.0
            The function name has been changed from ``derived_quantity`` to ``derived_array``. The old name is
            still available but will be removed in a future version.
        """
        if cls not in SimSnap._derived_array_registry:
            SimSnap._derived_array_registry[cls] = {}
        SimSnap._derived_array_registry[cls][fn.__name__] = fn
        fn.__stable__ = False
        cls._add_derived_to_doc(fn)

        return fn

    @classmethod
    def _add_derived_to_doc(cls, fn):
        if fn.__doc__ is None:
            return
        if cls is not SimSnap:
            if fn.__doc__.startswith("Derived ["):
                fn.__doc__ = f"Derived [{cls.__name__}, " + fn.__doc__[9:]
            else:
                fn.__doc__ = f"Derived [{cls.__name__}]: " + fn.__doc__
        else:
            if not fn.__doc__.startswith("Derived"):
                fn.__doc__ = "Derived: " + fn.__doc__

    @classmethod
    def stable_derived_array(cls, fn):
        """Function decorator to register a stable derivable quantity for all SimSnaps, or for a specific subclass.

        A stable derived array is one that is not expected to change over the course of a simulation, and so is
        only ever calculated once. Otherwise, it behaves just like a derived array; see
        :func:`~pynbody.snapshot.simsnap.SimSnap.derived_array`.

        If the function has a docstring, it will be prepended with "Derived: " to indicate that it is a derived
        quantity. If the derived array is being added for a specific subclass, a list of subclasses is also
        included in the automatic modifications to the docstring.

        For more information about derived quantities see :ref:`derived-quantities`, and specifically the
        subsection :ref:`stable-derived-quantities`.

        .. versionchanged:: 2.0
            The function name has been changed from ``stable_derived_quantity`` to ``stable_derived_array``. The old
            name is still available but will be removed in a future version.
        """

        if cls not in SimSnap._derived_array_registry:
            SimSnap._derived_array_registry[cls] = {}
        SimSnap._derived_array_registry[cls][fn.__name__] = fn
        fn.__stable__ = True
        cls._add_derived_to_doc(fn)
        return fn



    def _find_deriving_function(self, name):
        for cl in type(self).__mro__:
            if cl in self._derived_array_registry \
                    and name in self._derived_array_registry[cl]:
                return self._derived_array_registry[cl][name]
        else:
            return None

    def _derive_array(self, name, fam=None):
        """Calculate and store, for this SnapShot, the derivable array 'name'.
        If *fam* is not None, derive only for the specified family.

        This searches the registry of @X.derived_array functions
        for all X in the inheritance path of the current class.
        """
        global config

        calculated = False
        fn = self._find_deriving_function(name)
        if fn:
            logger.info("Deriving array %s" % name)
            with self.auto_propagate_off:
                if fam is None:
                    result = fn(self)
                    ndim = result.shape[-1] if len(
                        result.shape) > 1 else 1
                    self._create_array(
                        name, ndim, dtype=result.dtype, derived=not fn.__stable__)
                    write_array = self._get_array(
                        name, always_writable=True)
                else:
                    result = fn(self[fam])
                    ndim = result.shape[-1] if len(
                        result.shape) > 1 else 1

                    # check if a family array already exists with a different dtype
                    # if so, cast the result to the existing dtype
                    # numpy version < 1.7 does not support doing this in-place

                    if self._get_preferred_dtype(name) != result.dtype \
                       and self._get_preferred_dtype(name) is not None:
                        if int(np.version.version.split('.')[1]) > 6 :
                            result = result.astype(self._get_preferred_dtype(name),copy=False)
                        else :
                            result = result.astype(self._get_preferred_dtype(name))

                    self[fam]._create_array(
                        name, ndim, dtype=result.dtype, derived=not fn.__stable__)
                    write_array = self[fam]._get_array(
                        name, always_writable=True)

                self.ancestor._autoconvert_array_unit(result)

                write_array[:] = result
                if units.has_units(result):
                    write_array.units = result.units





    def _dirty(self, name):
        """Declare a given array as changed, so deleting any derived
        quantities which depend on it"""

        name = self._array_name_1D_to_ND(name) or name
        if name=='pos':
            for v in self.ancestor._persistent_objects.values():
                if 'kdtree' in v:
                    del v['kdtree']

        if not self.auto_propagate_off:
            for d_ar in self._dependency_tracker.get_dependents(name):
                if d_ar in self:
                    if self.is_derived_array(d_ar):
                        del self[d_ar]
                        self._dirty(d_ar)
                elif self.has_family_key(d_ar):
                    was_derived = False
                    for fam in self.families():
                        if self.is_derived_array(d_ar, fam):
                            was_derived = True
                            del self[fam][d_ar]
                    if was_derived:
                        self._dirty(d_ar)


    def is_derived_array(self, name, fam=None) -> bool:
        """Returns True if the array or family array of given name is
        auto-derived (and therefore read-only)."""
        fam = fam or self._unifamily
        if fam:
            return (name in self._family_derived_array_names[fam]) or name in self._derived_array_names
        elif name in list(self.keys()):
            return name in self._derived_array_names
        elif name in self.family_keys():
            return all([name in self._family_derived_array_names[i] for i in self._family_arrays[name]])
        else:
            return False

    def unlink_array(self, name):
        """If the named array is auto-derived, this destroys the link so that
        the array becomes editable but no longer auto-updates."""

        if self.is_derived_array(name):
            if name in self.family_keys():
                for fam in self._family_arrays[name]:
                    track = self._family_derived_array_names[fam]

                    if name in track:
                        del track[track.index(name)]

            else:
                del self._derived_array_names[
                    self._derived_array_names.index(name)]

        else:
            raise RuntimeError("Not a derived array")

    ############################################
    # CONVENIENCE FUNCTIONS
    ############################################
    def mean_by_mass(self, name) -> float | units.UnitBase:
        """Calculate the mean by mass of the specified array."""
        with self.immediate_mode:
            mass = self['mass']
            array = self[name]
            mass_indexing = (slice(None), ) + (np.newaxis, ) * (len(array.shape) - 1)
            return (mass[mass_indexing] * array).sum(axis=0) / mass.sum(axis=0)

    ############################################
    # SNAPSHOT DECORATION
    ############################################

    @classmethod
    def decorator(cl, fn):
        """Register a decorator to be applied to all instances of a given class.

        Decorators are called when a new instance of the class is created, and are passed the new instance as an
        argument."""
        if cl not in SimSnap._decorator_registry:
            SimSnap._decorator_registry[cl] = []
        SimSnap._decorator_registry[cl].append(fn)
        return fn

    def _decorate(self):
        for cl in type(self).__mro__:
            if cl in self._decorator_registry:
                for fn in self._decorator_registry[cl]:
                    fn(self)

    ############################################
    # KD-Tree
    ############################################

    def build_tree(self, num_threads=None, shared_mem=None) -> None:
        """Build a kdtree for SPH operations and for accelerating geometrical filters

        Parameters
        ----------

        num_threads : int, optional
            Number of threads to use for building and most operations on the tree.
            If None, the default number of threads is used.
        shared_mem : bool, optional
            Whether to use shared memory for the tree. This is used by the tangos library to
            share a kdtree between different processes. It is not recommended for general use,
            and defaults to False.
        """
        if not hasattr(self, 'kdtree'):
            from .. import kdtree
            from ..configuration import config
            boxsize = self._get_boxsize_for_kdtree()
            self.kdtree = kdtree.KDTree(self['pos'], self['mass'],
                                        leafsize=config['sph']['tree-leafsize'],
                                        boxsize=boxsize, num_threads=num_threads,
                                        shared_mem=shared_mem)

    def import_tree(self, serialized_tree, num_threads=None) -> None:
        """Import a precomputed kdtree from a serialized form.

        Throws ValueError if the tree is not compatible with the snapshot (though
        does not protect against all possible incompatibilities, so use with care)."""
        from .. import kdtree
        self.kdtree = kdtree.KDTree.deserialize(self['pos'], self['mass'],
                                         serialized_tree,
                                         boxsize=self._get_boxsize_for_kdtree(),
                                         num_threads=num_threads)
    def _get_boxsize_for_kdtree(self):
        boxsize = self.properties.get('boxsize', None)
        if boxsize:
            if units.is_unit_like(boxsize):
                boxsize = float(boxsize.in_units(self['pos'].units))
        else:
            boxsize = -1.0  # represents infinite box
        return boxsize


    ############################################
    # HASHING AND EQUALITY TESTING
    ############################################

    @property
    def _inclusion_hash(self):
        try:
            rval = self.__inclusion_hash
        except AttributeError:
            try:
                index_list = self.get_index_list(self.ancestor)
                hash = hashlib.md5(index_list.data)
                self.__inclusion_hash = hash.digest()
            except Exception:
                logging.warn(
                    "Encountered a problem while calculating your inclusion hash. %s" % traceback.format_exc())
            rval = self.__inclusion_hash
        return rval

    def __hash__(self):
        return hash((object.__hash__(self.ancestor), self._inclusion_hash))

    def __eq__(self, other):
        """Equality test for Snapshots. Returns true if both sides of the
        == operator point to the same data."""

        if self is other:
            return True
        return hash(self) == hash(other)

    ############################################
    # COPYING
    ############################################
    def __deepcopy__(self, memo=None):
        from .. import new

        create_args = {}
        for fam in family._registry:
            sl = self._get_family_slice(fam)
            if sl.start != sl.stop:
                create_args[fam.name] = len(self[fam])

        new_snap = new(**create_args)

        # ordering fix
        for k in copy.copy(list(new_snap._family_slice.keys())):
            new_snap._family_slice[k] = copy.copy(self._get_family_slice(k))

        for k in list(self.keys()):
            new_snap[k] = self[k]

        for k in self.family_keys():
            for fam in family._registry:
                if len(self[fam]) > 0:
                    self_fam = self[fam]
                    if k in list(self_fam.keys()) and not self_fam.is_derived_array(k):
                        new_snap[fam][k] = self_fam[k]

        new_snap.properties = copy.deepcopy(self.properties, memo)
        new_snap._file_units_system = copy.deepcopy(self._file_units_system, memo)

        return new_snap

    def get_copy_on_access_simsnap(self) -> SimSnap:
        """Return a new SimSnap that copies data out of this one when accessed

        This provides a degree of isolation (e.g. modifications made to the arrays in the copy-on-access
        view are not reflected back into the original snapshot). It is intended for advanced use, e.g. by
        tangos."""
        from .copy_on_access import CopyOnAccessSimSnap
        return CopyOnAccessSimSnap(self)

SimSnap.stable_derived_quantity = util.deprecated(SimSnap.stable_derived_array,
                                              "stable_derived_quantity has been renamed to stable_derived_array")

SimSnap.stable_derived_quantity.__doc__ = "Deprecated alias for :meth:`stable_derived_array`"

SimSnap.derived_quantity = util.deprecated(SimSnap.derived_array,
                                           "derived_quantity has been renamed to derived_array")

SimSnap.derived_quantity.__doc__ = "Deprecated alias for :meth:`derived_array`"



#
# pynbody documentation build configuration file, created by
# sphinx-quickstart on Mon Oct  3 11:57:24 2011.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

import numpy as np

#import sphinx_bootstrap_theme

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../pynbody'))
sys.path.append(os.path.abspath('sphinxext'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.todo',
              'sphinx.ext.napoleon',
              'sphinx.ext.coverage',
              'sphinx.ext.mathjax',
              'sphinx.ext.ifconfig',
              'sphinx.ext.viewcode',
              'matplotlib.sphinxext.mathmpl',
              'matplotlib.sphinxext.plot_directive',
              'sphinx.ext.inheritance_diagram',
              'sphinx_copybutton',
              'numpydoc',
              'nbsphinx'
              ]

nbsphinx_input_prompt = 'In [%s]:'
nbsphinx_output_prompt = 'Out[%s]:'
nbsphinx_execute = 'never' # the notebook is expensive to evaluate, so we need it pre-evaluated

autosummary_generate = True

ipython_warning_is_error = False
ipython_savefig_dir = 'plots'

# for .. plot:: directives
plot_working_directory = '.'


extensions+=['IPython.sphinxext.ipython_console_highlighting',
             'IPython.sphinxext.ipython_directive']


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

globaltoc_collapse = True

# General information about the project.
project = 'pynbody'
copyright = '2011-24, pynbody team'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
import pynbody

version = ".".join(pynbody.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = pynbody.__version__

# make sure plot module is imported for documenting
import pynbody.plot

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []



# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_book_theme'

# set up bootstrap
html_theme_options = {
    "repository_url": "https://github.com/pynbody/pynbody",
    "use_repository_button": True,
                       }

html_logo = "_static/logo.svg"

"""html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

html_theme_options = {
    'bootswatch_theme': 'spacelab',

    'navbar_site_name': 'Sections',

    'source_link_position': 'footer'
}"""

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = "_static/logo.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {'**': ['logo.html', 'fulltoc.html', 'relations.html', 'searchbox.html']}

# html_sidebars = {
#     '**': [
#         'about.html',
#         'navigation.html',
#         'relations.html',
#         'searchbox.html',
#     ]
# }

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'pynbodydoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'pynbody.tex', 'pynbody Documentation',
   'pynbody team (https://github.com/pynbody/pynbody/graphs/contributors)', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Additional stuff for the LaTeX preamble.
#latex_preamble = ''

# Documents to append as an appendix to all manuals.
#latex_appendices = []

#latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'pynbody', 'pynbody Documentation',
     ['pynbody team'], 1)
]

#autodoc_member_order = 'alphabetical'
#autoclass_content = 'both'

copybutton_copy_empty_lines = False
copybutton_selector =  "div.highlight > pre"
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.{3,5}: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_only_copy_prompt_lines = True

def hide_numpy_methods(app, what, name, obj, skip, options):
    if hasattr(obj, "__qualname__"):
        qname = obj.__qualname__
        if "SimArray" in qname:
            method_name = qname.split(".")[-1]
            if method_name in np.ndarray.__dict__:
                print("Skipping ", qname)
                return True
def setup(app):
    app.connect('autodoc-skip-member', hide_numpy_methods)

import matplotlib

matplotlib.rcParams['savefig.dpi'] = 200
matplotlib.rcParams['savefig.bbox'] = 'tight'
matplotlib.rcParams['savefig.pad_inches'] = 0.15



import numpy as np
import numpy.testing as npt
import pytest

import pynbody


@pytest.fixture
def dummy_file():
    # create a dummy gadget file
    f = pynbody.new(dm=2097152)
    f['iord'] = np.arange(2097152)
    f.properties['z']=1.6591479493605812
    f._filename = "testdata/rockstar/snapshot_015"

    return f

@pytest.fixture
def rockstar_halos(dummy_file):
    h = pynbody.halo.rockstar.RockstarCatalogue(dummy_file)
    return h

def test_rockstar_single_cpu():
    test = pynbody.halo.rockstar._RockstarCatalogueOneCpu("testdata/rockstar/halos_15.1.bin")
    assert test.halo_min_inclusive == 668
    assert test.halo_max_exclusive == 1426
    assert len(test) == 758
    assert (test.read_iords_for_halo(669) ==
        np.array([729699, 746082, 729443, 745954, 745827, 746083, 778725, 778981,
           729444, 762467, 762212, 778852, 729827, 778979, 746212, 746210,
           729572, 778726, 762595, 762211, 746085, 746086, 762469, 746213,
           745955, 762339, 729701, 778982, 762597, 762468, 778980, 746084,
           762341, 729571, 762596, 729829, 729700, 746211, 745956, 762342,
           729828, 762470, 762340, 745958, 778854, 762598, 745957, 778853,
           746214, 729698, 729570, 778724, 762594, 729573, 762338, 762466,
           778723, 778851])).all()

    props = test.read_properties_for_halo(669)

    assert props['id'] == 669
    assert np.allclose(props['Xoff'], 26.437117)
    assert np.allclose(props['vel'], np.array([  77.035904, -119.406364,  -27.567175]))

def test_load_rockstar(dummy_file):
    h = pynbody.halo.rockstar.RockstarCatalogue(dummy_file)
    assert len(h)==5851
    assert isinstance(h, pynbody.halo.rockstar.RockstarCatalogue)

def test_autodetect_rockstar_from_filename(dummy_file):
    dummy_file._filename = ""
    h = dummy_file.halos(filename="testdata/rockstar/halos_15.0.bin")
    assert isinstance(h, pynbody.halo.rockstar.RockstarCatalogue)

def test_rockstar_properties(rockstar_halos):
    h_properties = rockstar_halos[4977].properties
    assert h_properties['num_p']==40
    npt.assert_allclose(h_properties['pos'], [43.892704, 0.197397, 40.751919], rtol=1e-6)

def test_rockstar_all_properties(rockstar_halos):
    # rockstar reader doesn't generate units
    all_properties = rockstar_halos.get_properties_all_halos(with_units=False)
    assert len(all_properties['num_p'])==5851
    halo_index = rockstar_halos.number_mapper.number_to_index(4977)
    # halo_index will probably be 4977 but we shouldn't/won't take that for granted
    assert all_properties['num_p'][halo_index]==40

    npt.assert_allclose(all_properties['pos'][halo_index], [43.892704, 0.197397, 40.751919], rtol=1e-6)

@pytest.mark.parametrize('load_all', [True, False])
def test_rockstar_particles(rockstar_halos, load_all):
    if load_all:
        rockstar_halos.load_all()
    assert (np.sort(rockstar_halos[4977]['iord'])==[1801964, 1802346, 1818475, 1818729, 1818730, 1818857, 1818858, 1818859, 1818986,
                             1834860, 1834986, 1834987, 1835113, 1835114, 1835115, 1835116, 1835242, 1835243,
                             1835244, 1835369, 1835370, 1835371, 1835498, 1835499, 1851372, 1851625, 1851626,
                             1851627, 1851628, 1851754, 1851755, 1851756, 1851884, 1868010, 1868011, 1868012,
                             1884394, 1884395, 1900651, 1933291]).all()

def test_reject_unsuitable_rockstar_files():
    fwrong = pynbody.new(dm=2097152)
    fwrong.properties['z']=0
    with pytest.raises(RuntimeError):
        hwrong = fwrong.halos()
